import concurrent
import contextlib
import csv
import functools
import hashlib
import io
import json
import logging
import os
import re
import shutil
import uuid
from collections import defaultdict
from dataclasses import asdict, dataclass
from keyword import iskeyword
from pathlib import Path
from typing import IO, Any, Callable, Dict, List, Optional, Union

import datasets
import jinja2
import pandas as pd
import requests
import yaml
from typeguard import typechecked

from gantry.api_client import APIClient
from gantry.dataset import constants
from gantry.exceptions import (
    DataSchemaMismatchError,
    DatasetCommitNotFoundError,
    DatasetDeletedException,
    DatasetHeadOutOfDateException,
    DatasetNotFoundError,
    GantryRequestException,
    NoTabularDataError,
)
from gantry.query.core.dataframe import GantryDataFrame
from gantry.utils import (
    _download_file_to_IO,
    _upload_file_from_IO,
    download_file_from_url,
    get_files_checksum,
    get_s3_sha256_checksum,
    list_all_files,
    parse_s3_path,
    read_lines_from_url,
    upload_file_to_url,
)

logger = logging.getLogger(__name__)

BATCH_SIZE = 5


def _ensure_not_deleted(func: Callable) -> Callable:
    """
    Decorator to ensure we do not perform an operation on a deleted dataset, and instead warn the
    user that the dataset is deleted.
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        resp = self._api_client.request(
            "GET",
            f"/api/v1/datasets/{self.dataset_name}",
            raise_for_status=True,
        )
        dataset_info = resp["data"]

        if dataset_info["disabled"]:
            raise DatasetDeletedException("This dataset has been deleted!")

        return func(self, *args, **kwargs)

    return wrapper


def _ensure_dataset_exists_locally(func: Callable) -> Callable:
    """
    Decorator to ensure dataset exists locally, if not will warn the user to download the dataset.
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if not Path(self._get_file_path(constants.DATASET_MANIFEST_FILE)).exists():
            raise DatasetNotFoundError(
                f"Can't find dataset {self.dataset_name} locally, please run pull()"
                f"to download the dataset into your local workspace."
            )

        return func(self, *args, **kwargs)

    return wrapper


def _warn_when_local_changes(func: Callable) -> Callable:
    """
    Decorator to ensure there are no local changes in the dataset. If there are, will return None
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        # only do the following checking when HEAD file exist(dataset exists on local host)
        if os.path.exists(self._get_file_path(constants.DATASET_HEAD_FILE)):
            localdiff = self.get_diff()
            if (
                localdiff[constants.DELETED_FILES]
                or localdiff[constants.MODIFIED_FILES]
                or localdiff[constants.NEW_FILES]
            ) and not kwargs.get("forced", False):
                logger.warning(
                    "Local changes detected! Please run stash() to stash your local changes, "
                    "or run this function again with forced=True"
                )
                return None

        return func(self, *args, **kwargs)

    return wrapper


@dataclass
class DatasetFileInfo:
    file_name: str
    url: str
    version_id: Optional[str] = None
    sha256: Optional[str] = None

    def to_jsonl(self):
        if not self.file_name or not self.url or not self.version_id or not self.sha256:
            raise ValueError(
                f"Failed to create file info jsonl: {json.dumps(asdict(self))}. \
                    Incomplete dataset file info line! "
            )

        return f"{json.dumps(asdict(self))}\n"


class GantryDataset:
    """
    A class representing a lightweight container around your data that
    provides simple version control semantics for helping to make ML
    pipelines reproducible.
    """

    def __init__(
        self,
        api_client: APIClient,
        dataset_name: str,
        dataset_id: uuid.UUID,
        bucket_name: str,
        aws_region: str,
        dataset_s3_prefix: str,
        workspace: str,
    ):
        """
        :meta private:
        """
        self.dataset_name = dataset_name
        self.workspace = workspace
        self._api_client = api_client
        self._dataset_id = dataset_id
        self._bucket_name = bucket_name
        self._dataset_s3_prefix = dataset_s3_prefix
        self._aws_region = aws_region

    def list_versions(self) -> List[Dict[str, Any]]:
        """
        This method will list all the versions of the dataset. Each dataset version is a snapshot
        of the dataset at a particular point in time. The result will be sorted from latest
        to earliest.

        Example usage:

            >>> import gantry.dataset as gdataset
            >>> dataset = gdataset.get_dataset(dataset_name)
            >>> dataset.list_versions()
            [{'version_id': '7f100fca-f080-4daf-82d1-575e34234930',
              'dataset': 'dataset_name',
              'message': 'version notes',
              'created_at': 'Thu, 16 Feb 2023 00:16:33 GMT',
              'created_by': '04c894c7-a853-4fec-a024-66f5aae07b06',
              'is_latest_version': True},
             {'version_id': '300cdb42-a8c9-4b6e-a16d-4f6c925eb25d',
              'dataset': 'dataset_name',
              'message': 'version notes',
              'created_at': 'Thu, 16 Feb 2023 00:16:18 GMT',
              'created_by': '04c894c7-a853-4fec-a024-66f5aae07b06',
              'is_latest_version': False},
             {'version_id': '51e7771d-5ba9-4bdf-ab61-6d6ec9f7e066',
              'dataset': 'dataset_name',
              'message': 'initial dataset commit',
              'created_at': 'Thu, 16 Feb 2023 00:14:34 GMT',
              'created_by': '04c894c7-a853-4fec-a024-66f5aae07b06',
              'is_latest_version': False}]

        Returns:
            ``List[Dict[str, Any]]``: dataset versions from latest to earliest.
        """
        response = self._api_client.request(
            "GET", f"/api/v1/datasets/{self._dataset_id}/commits", raise_for_status=True
        )
        return [self._prune_commit_info(commit_info) for commit_info in response["data"]]

    @typechecked
    @_ensure_not_deleted
    @_warn_when_local_changes
    def pull(
        self, version_id: Optional[Union[str, uuid.UUID]] = None, forced: bool = False
    ) -> Optional[Dict[str, Any]]:
        """
        Pull a spicific version of the dataset from dataset server to local working directory.

        If version ID is provided, this method will pull the specific snapshot of your dataset
        based on the version id to working directory. If version ID is not provided, this method
        will pull the latest version. If forced is set to True, any local changes will be discarded.
        This method will only change the files in your local dataset folder, pull an older version
        will not affect the version history.

        Example usage:

            >>> import gantry.dataset as gdataset
            >>> dataset = gdataset.get_dataset(dataset_name)
            # pull the latest version of your dataset
            >>> dataset.pull()
            # pull a specific version of your dataset
            >>> dataset.pull("b787034a-798b-4bb3-a726-0e197ddb8aff")


        Args:
            version_id (str, optional): target version ID, defaults to ``None`` which will pull the
                latest version.
            forced (bool): whether to discard local changes or not when pulling,
                defaults to ``False``.

        Returns:
            Optional[Dict[str, Any]]: A dictionary with metadata representing the versioned rolled
            back to if successful, ``None`` otherwise.
        """
        target_commit = (
            self._get_commit(commit_id=version_id) if version_id else self._get_latest_commit()
        )

        metadata_version_id = target_commit[constants.METADATA_S3_FILE_VERSION]

        self._download_files(
            [
                DatasetFileInfo(
                    file_name=constants.DATASET_MANIFEST_FILE,
                    url=self._get_s3_url(constants.DATASET_MANIFEST_FILE),
                    version_id=metadata_version_id,
                )
            ]
        )

        diff = self._get_diff_for_pull()
        self._download_files(diff[constants.MODIFIED_FILES])  # Overwrite modified files
        self._download_files(diff[constants.DELETED_FILES])  # redownload deleted files

        # delete new added files
        for f in diff[constants.NEW_FILES]:
            os.remove(self._get_file_path(f.file_name))

        self._update_head(target_commit)
        if not os.path.exists(self._get_file_path(constants.STASH_FOLDER)):
            os.mkdir(self._get_file_path(constants.STASH_FOLDER))

        # Create the tabular_manifests and artifact folder
        (Path(self.workspace) / self.dataset_name / constants.TABULAR_MANIFESTS).resolve().mkdir(
            parents=True, exist_ok=True
        )

        (Path(self.workspace) / self.dataset_name / constants.ARTIFACTS).resolve().mkdir(
            parents=True, exist_ok=True
        )

        return self._prune_commit_info(target_commit)

    @_ensure_dataset_exists_locally
    def get_diff(self) -> Dict[str, List[str]]:
        """
        Show the local changes that have not been pushed to the server yet. This method will return
        a dictionary with three keys: ``new_files``, ``modified_files``, and ``deleted_files``.
        Each key will have a list of files that have been added, modified, or deleted respectively.

        Example usage:

            >>> import gantry.dataset as gdataset
            >>> dataset = gdataset.get_dataset(dataset_name)
            >>> dataset.pull()
            # make some changes to the dataset
            >>> dataset.get_diff()
            {'new_files': ['tabular_manifests/new_file.csv', 'artifacts/image.png'],
             'modified_files': ['tabular_manifests/modified_file_1.csv'],
             'deleted_files': ['tabular_manifests/deleted_file.csv']}

        Returns:
            ``dict``: a dictionary representing the diff, which looks like this:

            .. code-block:: json

                {
                    "new_files": List[str],
                    "modified_files": List[str],
                    "deleted_files": List[str],
                }
        """

        diff = self._get_diff()
        return {
            constants.NEW_FILES: [f for f, _ in diff[constants.NEW_FILES]],
            constants.MODIFIED_FILES: [f for f, _ in diff[constants.MODIFIED_FILES]],
            constants.DELETED_FILES: [f for f, _ in diff[constants.DELETED_FILES]],
        }

    @typechecked
    @_ensure_dataset_exists_locally
    @_ensure_not_deleted
    def push_version(self, message: str) -> Dict[str, Any]:
        """
        This method will create a new dataset version in the remote server with a version
        message. The new version will include all the local changes that have not been pushed.
        If there are no local changes, this method will return the current version.

        To avoid race conditions, this method will check that the HEAD of the dataset is up to date.
        If the HEAD is not up to date, this method will raise a ``DatasetHeadOutOfDateException``.
        In that case you could stash your local change and pull the latest version of the dataset
        and try again. Check out the documents for ``stash()``, ``pull()`` and ``restore()`` for
        more details.

        Example usage:

            >>> import gantry.dataset as gdataset
            >>> dataset = gdataset.get_dataset(dataset_name)
            >>> dataset.pull()
            # make some changes to the dataset
            >>> dataset.get_diff()
            {'new_files': ['tabular_manifests/new_file.csv', 'artifacts/image.png'],
            'modified_files': ['tabular_manifests/modified_file_1.csv'],
            'deleted_files': ['tabular_manifests/deleted_file.csv']}
            >>> dataset.push_version("new version notes")
            {'version_id': '09575ee7-0407-44b8-ae88-765a8270b17a',
            'dataset': 'dataset_name',
            'message': 'new version notes',
            'created_at': 'Wed, 15 Feb 2023 22:17:55 GMT',
            'created_by': '04c894c7-a853-4fec-a024-66f5aae07b06',
            'is_latest_version': True}
            # After pushing the changes, the diff should be empty
            >>> dataset.get_diff()
            {'new_files':[],
            'modified_files':[],
            'deleted_files':[]}


        Args:
            message (str): a non-empty string representing a message to associate with
                the created version.
        Returns:
            ``Dict[str, Any]``: a dictionary with metadata representing the version created.
        """
        commit_id = uuid.uuid4()
        diff = self._get_diff_for_commit(commit_id)

        if (
            not diff[constants.NEW_FILES]
            and not diff[constants.MODIFIED_FILES]
            and not diff[constants.DELETED_FILES]
        ):
            logger.warning("No local changes to add!")
            return self._get_current_commit()

        # Check that HEAD is up to date
        resp = self._api_client.request(
            "GET", f"/api/v1/datasets/{self._dataset_id}/commits", raise_for_status=True
        )
        latest_commit_id = resp["data"][0]["id"]

        with open(self._get_file_path(constants.DATASET_HEAD_FILE), "r") as f:
            current_commit = json.load(f)

            if current_commit["id"] != latest_commit_id:
                raise DatasetHeadOutOfDateException

        diff[constants.NEW_FILES] = self._upload_files(diff[constants.NEW_FILES])
        diff[constants.MODIFIED_FILES] = self._upload_files(diff[constants.MODIFIED_FILES])

        self._backup_file(constants.DATASET_MANIFEST_FILE)
        self._update_manifest_file(diff)

        result = self._upload_files(
            [
                DatasetFileInfo(
                    file_name=constants.DATASET_MANIFEST_FILE,
                    url=self._get_s3_url(constants.DATASET_MANIFEST_FILE),
                )
            ]
        )
        version_id = result[0].version_id
        try:
            resp = self._api_client.request(
                "POST",
                f"/api/v1/datasets/{self._dataset_id}/commits",
                json={
                    "message": message,
                    "metadata_s3_file_version": version_id,
                    "parent_commit_id": self._get_current_commit()["id"],
                    "commit_id": commit_id,
                },
                raise_for_status=True,
            )
            os.remove(
                self._get_file_path(f"{constants.DATASET_MANIFEST_FILE}{constants.BACKUP_SUFFIX}")
            )
        except GantryRequestException as e:
            self._recover_file(constants.DATASET_MANIFEST_FILE)
            if (
                e.status_code == 400
                and 'Details: {"error":"Parent version out of date!","response":"error"}'
                in e.args[0]
            ):
                raise DatasetHeadOutOfDateException
            raise e

        commit_info = resp["data"]
        self._update_head(commit_info)

        return self._prune_commit_info(commit_info)

    def push_gantry_dataframe(self, gantrydf: GantryDataFrame, filename: str = ""):
        """
        This method will take a GantryDataFrame as input, and save its contents to a new file in
        your dataset's ``tabular_manifests`` folder, and create a new version. Note that all these
        operations happen in the remote server, if you have your dataset checked out locally,
        please make sure you call ``pull()`` after this operation to get the latest data.

        Example usage:

            >>> import gantry.dataset as gdataset
            >>> dataset = gdataset.get_dataset(dataset_name)
            >>> dataset.pull()
            >>> os.listdir(workspace + dataset_name)
            ['.dataset_metadata', 'dataset_config.yaml', 'README.md', 'tabular_manifests']
            >>> os.listdir(workspace + dataset_name + '/tabular_manifests')
            []
            >>> gantrydf = gquery.query(
            ...   application=application,
            ...   start_time=start,
            ...   end_time=end,
            ...   version="1.1",
            ... )
            >>>
            >>> _ = dataset.push_gantry_dataframe(gantrydf)
            >>>
            >>> _ = dataset.pull()
            >>> os.listdir(workspace + dataset_name + '/tabular_manifests')
            ['GantryDataAutoSave__4390b343-b802-4add-ab51-f86e53979c73_2023-02-13_12:09_to_2023-02-13_17:09_rows_0_69.csv']

        Args:
            gantrydf (GantryDataFrame): the GantryDataFrame that needs to be added to the dataset.
            filename (str): the name to be given to the file with data from the GantryDataFrame.
        Returns:
            Latest commit info.
        """
        start_time = gantrydf.query_info.start_time
        end_time = gantrydf.query_info.end_time
        final_filename = filename
        if filename == "":
            final_filename = f"GantryDataAutoSave__{uuid.uuid4()}_{start_time.strftime('%Y-%m-%d_%H:%M')}_to_{end_time.strftime('%Y-%m-%d_%H:%M')}"  # noqa: E501
        app_node_id = gantrydf.query_info.application_node_id
        filters = gantrydf.filters
        tags = gantrydf.tags
        resp = self._api_client.request(
            "POST",
            f"/api/v1/datasets/{self._dataset_id}/raw_data",
            json={
                "model_id": app_node_id,
                "raw_data_query_parameters": {
                    "filters": filters,
                    "tags": tags,
                    "include_join_id": True,
                    "hidden_columns": None,
                    "start_time": start_time,
                    "end_time": end_time,
                },
                "raw_data_csv_prefix": final_filename,
            },
            raise_for_status=True,
        )
        commit_info = resp["data"]
        return self._prune_commit_info(commit_info)

    @typechecked
    @_ensure_not_deleted
    def push_file(
        self,
        file: IO,
        filename: Optional[str] = None,
        message: Optional[str] = None,
        parent_version_id: Optional[Union[uuid.UUID, str]] = None,
    ) -> Dict[str, Any]:
        """
        This method will add a new file to the dataset in Gantry server and create a new version of
        the dataset. With this method you can easily add new files to the dataset without having to
        pull the dataset locally. If you have your dataset checked out locally, please make sure you
        call ``pull()`` after this operation to get the latest data. The difference between this
        method and ``push_tabular_file()`` is that this method can handle both tabular and
        non-tabular files. If you are uploading a tabular file, please use ``push_tabular_file()``
        instead.


        Args:
            file (IO): the file to be uploaded
            filename (str, Optional): the name of the file to be uploaded. Defaults to the
                name of the file passed in the ``file`` parameter.
            message (str, Optional): the version message that will be associated with the
                upload of the file. Defaults to a generic message if not set
            parent_version_id (uuid.UUID, Optional): If specified, SDK will check whether you are
                making changes on top of the latest version, if not the operation will fail.

        Returns:
            ``Dict[str, Any]``: A dictionary with metadata representing the version created.
        """
        if filename:
            name = filename
        else:
            name = os.path.basename(file.name)

        return self._add_single_file(file, name, message, parent_version_id)

    @typechecked
    @_ensure_not_deleted
    def push_tabular_file(
        self,
        file: IO,
        filename: Optional[str] = None,
        message: Optional[str] = None,
        parent_version_id: Optional[Union[uuid.UUID, str]] = None,
    ) -> Dict[str, Any]:
        """
        This method will add a tabular file to the dataset's ``tabular_manifests`` folder and create
        a new version of the dataset on dataset server. With this method you can easily add new
        tabular files to an existing dataset without having to pull the dataset locally. If you
        have your dataset checked out locally, please make sure you call ``pull()`` after this
        operation to get the latest data. If you are trying to add a non-tabular file, please
        use ``push_file()`` instead. Note: the tabular filename must be postfixed with ``.csv``.

        Example usage:

            >>> import gantry.dataset as gdataset
            >>> dataset = gdataset.get_dataset(dataset_name)
            >>> dataset.push_tabular_file(
            ...     open("{file_to_be_added}.csv"),
            ...     "{dataset_file_name}.csv",
            ...     "version info"
            ...    )
            {'version_id': '2b575ee7-0407-44b8-ae88-765a8270b17a',
            'dataset': 'dataset_name',
            'message': 'version info',
            'created_at': 'Wed, 15 Feb 2023 22:17:55 GMT',
            'created_by': '04c894c7-a853-4fec-a024-66f5aae07b06',
            'is_latest_version': True}

        Args:
            file (IO): the csv file to be uploaded
            filename (str, Optional): the name of the file to be uploaded. Defaults to the local
                file name of the file. Name must end with ``.csv``.
            message (str, optional): the version message that will be associated with the
                upload of the file. Defaults to a generic message if not set
            parent_version_id (uuid.UUID, optional): If specified, SDK will check whether you are
                making changes on top of the latest version, if not the operation will fail.

        Returns:
            ``Dict[str, Any]``: A dictionary with metadata representing the version created.
        """
        if filename:
            name = filename
        else:
            name = os.path.basename(file.name)

        if name[-4:] != constants.CSV_SUFFIX:
            raise ValueError(f"filename must end with '.csv'! Got filename: {name}")

        return self._add_single_file(file, name, message, parent_version_id)

    @typechecked
    @_ensure_not_deleted
    def push_dataframe(
        self,
        dataframe: pd.DataFrame,
        filename: Optional[str] = None,
        message: Optional[str] = None,
        parent_version_id: Optional[Union[uuid.UUID, str]] = None,
    ) -> Dict[str, Any]:
        """
        Add a dataframe to the dataset's ``tabular_manifests`` folder as a csv file and create a new
        version of the dataset on the dataset server. With this method you can easily add new a
        dataframe to an existing dataset without having to pull the dataset locally. If you have
        your dataset checked out locally, please make sure you call ``pull()`` after this operation.

        Example usage:

            >>> import gantry.dataset as gdataset
            >>> dataset = gdataset.get_dataset(dataset_name)
            >>> dataset.push_dataframe(
            ...     pd.DataFrame({"col1": [1, 2, 3], "col2": [4, 5, 6]}),
            ...     "dataframe.csv",
            ...     "version info"
            ... )
            {'version_id': 'fa575ee7-0407-44b8-ae88-765a8270b17a',
            'dataset': 'dataset_name',
            'message': 'version info',
            'created_at': 'Wed, 15 Feb 2023 22:17:55 GMT',
            'created_by': '04c894c7-a853-4fec-a024-66f5aae07b06',
            'is_latest_version': True}

        Args:
            dataframe (pd.DataFrame): the dataframe to be uploaded
            filename (str, Optional): the name for the csv that will be uploaded. Defaults to
                dataframe-<random-hash>.csv
            message (str, Optional): the version message that will be associated with the
                upload of the file. Defaults to a generic message if not set
            parent_version_id (uuid.UUID, Optional): If specified, SDK will check whether you are
                making changes on top of the latest version, if not the operation will fail.

        Returns:
            ``Dict[str, Any]``: A dictionary with metadata representing the version created.
        """
        if filename:
            if filename[-4:] == constants.CSV_SUFFIX:
                name = filename
            else:
                name = f"{filename}{constants.CSV_SUFFIX}"
        else:
            hash = hashlib.md5(pd.util.hash_pandas_object(dataframe).values).hexdigest()
            name = f"dataframe-{hash}{constants.CSV_SUFFIX}"

        file = io.StringIO()
        dataframe.to_csv(file, index=False)
        file.seek(0)

        return self._add_single_file(file, name, message, parent_version_id)

    @typechecked
    @_ensure_not_deleted
    @_warn_when_local_changes
    def rollback(
        self, version_id: Union[str, uuid.UUID], forced: bool = False
    ) -> Optional[Dict[str, Any]]:
        """
        Rollback the dataset data to a previous version and create a new dataset version based on
        the target version id. With this method you can easily rollback to any previous version of
        your dataset and continue working on top of it. Note that to do this, you must have your
        local dataset folder up to date. Also this method will automatically pull the latest version
        to local dataset folder after the rollback. If forced is set to True, any local changes
        will be discarded.

        Example usage:

            >>> import gantry.dataset as gdataset
            >>> dataset = gdataset.get_dataset(dataset_name)
            >>> dataset.pull()
            >>> dataset.list_versions()
            [{'version_id': 'd6f42bd0-b7a9-4089-88fb-be7cee6d4a83',
            'dataset': 'dataset_name',
            'message': 'version info',
            'created_at': 'Wed, 08 Feb 2023 22:02:02 GMT',
            'created_by': 'db459d6d-c83b-496d-b659-e48bca971156',
            'is_latest_version': True},
            {'version_id': 'b787034a-798b-4bb3-a726-0e197ddb8aff',
            'dataset': 'dataset_name',
            'message': 'initial dataset commit',
            'created_at': 'Wed, 08 Feb 2023 22:00:26 GMT',
            'created_by': 'db459d6d-c83b-496d-b659-e48bca971156',
            'is_latest_version': False}]
            >>> dataset.rollback("b787034a-798b-4bb3-a726-0e197ddb8aff")
            {'version_id': '23bc4d35-0df2-424c-9156-d5ca105eb4c1',
            'dataset': 'dataset_name',
            'message': 'Rollback dataset to version: b787034a-798b-4bb3-a726-0e197ddb8aff',
            'created_at': 'Thu, 09 Feb 2023 00:36:07 GMT',
            'created_by': 'db459d6d-c83b-496d-b659-e48bca971156',
            'is_latest_version': True}
            >>> dataset.list_versions()
            [{'version_id': '23bc4d35-0df2-424c-9156-d5ca105eb4c1',
            'dataset': 'dataset_name',
            'message': 'Rollback dataset to version: b787034a-798b-4bb3-a726-0e197ddb8aff',
            'created_at': 'Thu, 09 Feb 2023 00:36:07 GMT',
            'created_by': 'db459d6d-c83b-496d-b659-e48bca971156',
            'is_latest_version': True},
            {'version_id': 'd6f42bd0-b7a9-4089-88fb-be7cee6d4a83',
            'dataset': 'dataset_name',
            'message': 'version info',
            'created_at': 'Wed, 08 Feb 2023 22:02:02 GMT',
            'created_by': 'db459d6d-c83b-496d-b659-e48bca971156',
            'is_latest_version': False},
            {'version_id': 'b787034a-798b-4bb3-a726-0e197ddb8aff',
            'dataset': 'dataset_name',
            'message': 'initial dataset commit',
            'created_at': 'Wed, 08 Feb 2023 22:00:26 GMT',
            'created_by': 'db459d6d-c83b-496d-b659-e48bca971156',
            'is_latest_version': False}]

        Args:
            version_id (str): ID of the version to rollback to.
            forced (bool): whether to discard local changes when rolling back.

        Returns:
            Optional[Dict[str, Any]]: A dictionary with metadata representing the versioned
            rolled back to if successful, ``None`` otherwise.
        """
        target_commit = self._get_commit(version_id)

        resp = self._api_client.request(
            "POST",
            f"/api/v1/datasets/{self._dataset_id}/commits",
            json={
                "message": f"Rollback dataset to version: {target_commit['id']}",
                "metadata_s3_file_version": target_commit["metadata_s3_file_version"],
                "parent_commit_id": self._get_current_commit()["id"],
            },
            raise_for_status=True,
        )

        commit_info = resp["data"]
        self.pull(commit_info["id"], forced=forced)

        return self._prune_commit_info(commit_info)

    def stash(self) -> None:
        """
        This method will stash all the changes in your local dataset folder to a temporary folder.
        You can use this method to save your changes before the ``pull()`` method. Later on you can
        use the ``restore()`` method to restore your changes. Note that if you run ``stash()``
        multiple times, the changes will be overwritten and only the latest stash will be kept.

        Example usage:

            >>> import gantry.dataset as gdataset
            >>> dataset = gdataset.get_dataset(dataset_name)
            >>> dataset.pull()
            # make some changes to your dataset
            >>> dataset.get_diff()
            {'new_files': ['tabular_manifests/new_file.csv', 'artifacts/image.png'],
            'modified_files': ['tabular_manifests/modified_file_1.csv'],
            'deleted_files': ['tabular_manifests/deleted_file.csv']}
            >>> dataset.stash()
            >>> dataset.get_diff()
            {'new_files': [],
            'modified_files': [],
            'deleted_files': []}

        Returns:
            ``None``
        """
        # deletes existing stash
        if os.path.exists(self._get_file_path(constants.DATASET_STASH_FILE)):
            with open(self._get_file_path(constants.DATASET_STASH_FILE)) as stashfile:
                old_stash = json.load(stashfile)
            for file in old_stash[constants.MODIFIED_FILES] + old_stash[constants.NEW_FILES]:
                with contextlib.suppress(FileNotFoundError):  # file not exist is OK
                    os.remove(self._get_file_path(f"{constants.STASH_FOLDER}/{file}"))

        # creates new stash
        local_diff = self.get_diff()
        with open(self._get_file_path(constants.DATASET_STASH_FILE), "w+") as stashfile:
            json.dump(local_diff, stashfile)

        # stash files
        for file in local_diff[constants.MODIFIED_FILES]:
            self._stash_file(file)
        for file in local_diff[constants.NEW_FILES]:
            self._stash_file(file)

        # re-sync to current commit
        diff = self._get_diff_for_pull()
        logger.warning(diff)
        self._download_files(diff[constants.MODIFIED_FILES])  # overwrite modified files
        self._download_files(diff[constants.DELETED_FILES])  # redownload deleted files

        # delete new added files
        for file in diff[constants.NEW_FILES]:
            os.remove(self._get_file_path(file.file_name))

    def restore(self) -> None:
        """
        This method will restore the changes from the stash to your local dataset folder. Note that
        if you have made changes to your local dataset folder after the ``stash()`` method, those
        changes could be overwritten by this method.

        Example usage:

            >>> import gantry.dataset as gdataset
            >>> dataset = gdataset.get_dataset(dataset_name)
            >>> dataset.pull()
            # make some changes to your local dataset folder
            >>> dataset.get_diff()
            {'new_files': ['tabular_manifests/new_file.csv', 'artifacts/image.png'],
            'modified_files': ['tabular_manifests/modified_file_1.csv'],
            'deleted_files': ['tabular_manifests/deleted_file.csv']}
            >>> dataset.stash()
            >>> dataset.get_diff()
            {'new_files': [],
            'modified_files': [],
            'deleted_files': []}
            >>> dataset.restore()
            >>> dataset.get_diff()
            {'new_files': ['tabular_manifests/new_file.csv', 'artifacts/image.png'],
            'modified_files': ['tabular_manifests/modified_file_1.csv'],
            'deleted_files': ['tabular_manifests/deleted_file.csv']}

        Returns:
            ``None``
        """
        if not os.path.exists(self._get_file_path(constants.DATASET_STASH_FILE)):
            logger.warning("No stashed files were found!")
            return

        diff = {}
        with open(self._get_file_path(constants.DATASET_STASH_FILE)) as stashfile:
            diff = json.load(stashfile)
        os.remove(self._get_file_path(constants.DATASET_STASH_FILE))
        if not diff:
            return

        for file in diff[constants.MODIFIED_FILES]:
            self._restore_stash_file(file)
        for file in diff[constants.NEW_FILES]:
            self._restore_stash_file(file)
        for file in diff[constants.DELETED_FILES]:
            os.remove(self._get_file_path(file))

    def delete(self) -> None:
        """
        Mark the dataset as deleted. As for now this will only hide dataset from the list_datasets
        API results, it will not delete the dataset data from the server, we are working on
        implementing the hard delete feature.

        Returns:
            ``None``
        """
        self._api_client.request(
            "DELETE",
            f"/api/v1/datasets/{self._dataset_id}",
            raise_for_status=True,
        )
        logger.info(f"Dataset {self.dataset_name} has been deleted")
        return

    @typechecked
    def _get_commit(self, commit_id: Union[str, uuid.UUID]) -> Dict[str, Any]:
        """
        Get commit details

        Args:
            commit_id (str): commit id
        Returns:
            Commit information
        """
        try:
            resp = self._api_client.request(
                "GET",
                f"/api/v1/datasets/{self._dataset_id}/commits/{commit_id}",
                raise_for_status=True,
            )
        except GantryRequestException as e:
            if e.status_code == 404:
                raise DatasetCommitNotFoundError(
                    f"Could not find dataset commit with id {commit_id}"
                ) from e
            raise
        return resp["data"]

    def _get_latest_commit(self) -> Dict[str, Any]:
        """
        Get latest commit details

        Returns:
            Commit information
        """
        resp = self._api_client.request(
            "GET", f"/api/v1/datasets/{self._dataset_id}/commits", raise_for_status=True
        )
        return resp["data"][0]  # return latest commit

    def _update_manifest_file(self, diff: Dict[str, List[DatasetFileInfo]]):
        new_files = {f.file_name: f for f in diff[constants.NEW_FILES]}
        modified_files = {f.file_name: f for f in diff[constants.MODIFIED_FILES]}
        deleted_files = {f.file_name: f for f in diff[constants.DELETED_FILES]}

        with open(
            self._get_file_path(f"{constants.DATASET_MANIFEST_FILE}{constants.NEW_SUFFIX}"), "w"
        ) as new_gantry_manifest:
            with open(
                self._get_file_path(constants.DATASET_MANIFEST_FILE), "r"
            ) as cur_gantry_manifest:
                for line in cur_gantry_manifest:
                    cur_file_info = json.loads(line)
                    cur_file_name = cur_file_info[constants.FILE_NAME]

                    if cur_file_name in modified_files:  # update modified files
                        new_gantry_manifest.write(modified_files[cur_file_name].to_jsonl())
                    elif cur_file_name in deleted_files:  # remove deleted files
                        continue
                    else:  # keep unchanged files
                        new_gantry_manifest.write(line)

            for _, file_info in new_files.items():
                new_gantry_manifest.write(file_info.to_jsonl())  # add new added files

        os.replace(
            self._get_file_path(f"{constants.DATASET_MANIFEST_FILE}{constants.NEW_SUFFIX}"),
            self._get_file_path(constants.DATASET_MANIFEST_FILE),
        )

    def _get_diff_for_commit(self, commit_id: uuid.UUID):
        repo_diff = self._get_diff()
        commit_diff = {}

        # generate new url for new files
        commit_diff[constants.NEW_FILES] = [
            DatasetFileInfo(
                file_name=file_name,
                url=self._get_s3_url(file_name, commit_id),
                sha256=checksum,
            )
            for file_name, checksum in repo_diff[constants.NEW_FILES]
        ]
        # generate new url for modified files
        commit_diff[constants.MODIFIED_FILES] = [
            DatasetFileInfo(
                file_name=file_name,
                url=self._get_s3_url(file_name, commit_id),
                sha256=checksum,
            )
            for file_name, checksum in repo_diff[constants.MODIFIED_FILES]
        ]
        # don't need url for deleted files
        commit_diff[constants.DELETED_FILES] = [
            DatasetFileInfo(
                file_name=file_name,
                url="",
                sha256=checksum,
            )
            for file_name, checksum in repo_diff[constants.DELETED_FILES]
        ]

        return commit_diff

    def _get_diff_for_pull(self):
        repo_diff = self._get_diff()
        sync_diff = {}

        commit_snapshot = {}
        with open(self._get_file_path(constants.DATASET_MANIFEST_FILE)) as cur_gantry_manifest:
            for line in cur_gantry_manifest:
                file_info = json.loads(line)
                commit_snapshot[file_info[constants.FILE_NAME]] = file_info

        # don't need url for new files and use checksum for empty string
        sync_diff[constants.NEW_FILES] = [
            DatasetFileInfo(
                file_name=file_name,
                url="",
                sha256=constants.EMPTY_STR_SHA256,
            )
            for file_name, _ in repo_diff[constants.NEW_FILES]
        ]
        # retrieve url for modified files
        sync_diff[constants.MODIFIED_FILES] = [
            DatasetFileInfo(
                file_name=file_name,
                url=commit_snapshot[file_name][constants.URL],
                sha256=commit_snapshot[file_name][constants.SHA256],
                version_id=commit_snapshot[file_name][constants.VERSION_ID],
            )
            for file_name, _ in repo_diff[constants.MODIFIED_FILES]
        ]
        # retrieve url for deleted files
        sync_diff[constants.DELETED_FILES] = [
            DatasetFileInfo(
                file_name=file_name,
                url=commit_snapshot[file_name][constants.URL],
                sha256=commit_snapshot[file_name][constants.SHA256],
                version_id=commit_snapshot[file_name][constants.VERSION_ID],
            )
            for file_name, _ in repo_diff[constants.DELETED_FILES]
        ]

        return sync_diff

    def _get_diff(self):
        """
        Generate local diff.

        Args:
            return_local_info (bool): if true return local file info for modified files else
            return file info in dataset repo

        Returns:
            {
                "new_files": List[Tuple(str,str)],
                "modified_files": List[Tuple(str,str)],
                "deleted_files": List[Tuple(str,str)],
            }
        """
        repo_diff = defaultdict(list)

        commit_snapshot = {}
        with open(self._get_file_path(constants.DATASET_MANIFEST_FILE)) as cur_gantry_manifest:
            for line in cur_gantry_manifest:
                file_info = json.loads(line)
                commit_snapshot[file_info[constants.FILE_NAME]] = file_info

        local_files = get_files_checksum(Path(self.workspace) / self.dataset_name)

        # TODO:// Simplify the following logic to make it easier to maintain
        for file_name, checksum in local_files.items():
            if file_name.startswith(constants.GANTRY_FOLDER):  # skip .dataset_metadata folder
                continue
            elif file_name.startswith(constants.STASH_FOLDER):  # skip stash folder
                continue
            elif file_name not in commit_snapshot:  # new added file
                repo_diff[constants.NEW_FILES].append((file_name, checksum))
            elif checksum != commit_snapshot[file_name].get(
                constants.SHA256
            ):  # if file_name in current_snapshot --> check if it has been modified
                repo_diff[constants.MODIFIED_FILES].append((file_name, checksum))

        for file_name in commit_snapshot:
            if file_name not in local_files:
                repo_diff[constants.DELETED_FILES].append((file_name, constants.EMPTY_STR_SHA256))

        return repo_diff

    def _add_single_file(
        self,
        file: IO,
        filename: str,
        commit_msg: Optional[str],
        parent_commit_id: Optional[Union[uuid.UUID, str]],
    ):
        # check if parent commit is out of date
        if parent_commit_id is not None:
            if str(parent_commit_id) != self._get_latest_commit()["id"]:
                raise DatasetHeadOutOfDateException()

        if commit_msg is None:
            commit_msg = f"Uploaded {filename} to dataset"
        commit_id = uuid.uuid4()

        # get sha256 checksum
        try:
            file.seek(0)
            file_sha256 = get_s3_sha256_checksum(file.read())
        except TypeError:
            file.seek(0)
            file_sha256 = get_s3_sha256_checksum(bytes(file.read(), encoding="utf-8"))
        file.seek(0)

        # get file directory
        file_dir = constants.ARTIFACTS
        if len(filename) > 4 and filename[-4:] == constants.CSV_SUFFIX:
            file_dir = constants.TABULAR_MANIFESTS

        # get put urls:
        upload_file_info = DatasetFileInfo(
            file_name=f"{file_dir}/{filename}",
            url=self._get_s3_url(f"{file_dir}/{filename}", commit_id),
            sha256=file_sha256,
        )
        manifest_file_info = DatasetFileInfo(
            file_name=constants.DATASET_MANIFEST_FILE,
            url=self._get_s3_url(constants.DATASET_MANIFEST_FILE),
        )
        urls = self._generate_put_presign_url([upload_file_info, manifest_file_info])

        # upload added file to s3
        upload_file_info.version_id = _upload_file_from_IO(
            urls[self._get_obj_key(upload_file_info.url)], file
        )

        # download manifest file
        url = self._generate_get_presign_url([manifest_file_info])[
            self._get_obj_key(manifest_file_info.url)
        ]
        manifest_file = io.BytesIO()
        _download_file_to_IO(url, manifest_file)
        # update manifest file
        manifest_file.write(bytes(upload_file_info.to_jsonl(), encoding="utf-8"))
        manifest_file.seek(0)
        # upload new manifest file
        version_id = _upload_file_from_IO(
            urls[self._get_obj_key(manifest_file_info.url)], manifest_file
        )

        # create dataset commit
        try:
            resp = self._api_client.request(
                "POST",
                f"/api/v1/datasets/{self._dataset_id}/commits",
                json={
                    "message": commit_msg,
                    "metadata_s3_file_version": version_id,
                    "parent_commit_id": (
                        self._get_latest_commit()["id"]
                        if parent_commit_id is None
                        else parent_commit_id
                    ),
                    "commit_id": commit_id,
                },
                raise_for_status=True,
            )
        except GantryRequestException as e:
            if (
                e.status_code == 400
                and 'Details: {"error":"Parent version out of date!","response":"error"}'
                in e.args[0]
            ):
                raise DatasetHeadOutOfDateException
            raise e

        commit_info = resp["data"]
        return self._prune_commit_info(commit_info)

    def _get_file_path(self, file_name: str) -> str:
        return os.path.join(self.workspace, self.dataset_name, file_name)

    def _get_s3_url(self, file_name: str, commit_id: Optional[uuid.UUID] = None) -> str:
        if commit_id:
            return (
                f"s3://{self._bucket_name}/{self._dataset_s3_prefix}/"
                f"{self.dataset_name}/{commit_id}/{file_name}"
            )
        else:
            return (
                f"s3://{self._bucket_name}/{self._dataset_s3_prefix}/"
                f"{self.dataset_name}/{file_name}"
            )

    def _get_obj_key(self, s3_url: str) -> str:
        _, obj_key = parse_s3_path(s3_url)
        return obj_key

    def _update_head(self, commit_info: Dict[str, Any]):
        with open(self._get_file_path(constants.DATASET_HEAD_FILE), "w") as f:
            json.dump(commit_info, f)

    def _get_current_commit(self) -> Dict[str, Any]:
        with open(self._get_file_path(constants.DATASET_HEAD_FILE), "r") as f:
            return json.load(f)

    def _backup_file(self, file_path: str):
        # make a copy of the original file
        shutil.copyfile(
            self._get_file_path(file_path),
            self._get_file_path(f"{file_path}{constants.BACKUP_SUFFIX}"),
        )

    def _recover_file(self, file_path: str):
        # recover the file from a backup copy
        os.replace(
            self._get_file_path(f"{file_path}{constants.BACKUP_SUFFIX}"),
            self._get_file_path(file_path),
        )

    def _stash_file(self, file_path: str):
        # move file to stash folder
        dest_path = self._get_file_path(f"{constants.STASH_FOLDER}/{file_path}")
        os.makedirs(Path(dest_path).parent, exist_ok=True)
        os.replace(
            self._get_file_path(file_path),
            dest_path,
        )

    def _restore_stash_file(self, file_path: str):
        # recover file from stash folder
        dest_path = self._get_file_path(file_path)
        os.makedirs(Path(dest_path).parent, exist_ok=True)
        os.replace(
            self._get_file_path(f"{constants.STASH_FOLDER}/{file_path}"),
            dest_path,
        )

    def _download_files(self, file_list: List[DatasetFileInfo]):
        """
        download file to the specific path

        Args:
            file_list (List[DatasetFileInfo]):
        """
        if len(file_list) == 0:
            logger.info("No files to download")
            return
        url_args = []
        path_args = []
        presigned_urls = {}
        for i in range(0, len(file_list), BATCH_SIZE):
            presigned_urls.update(
                self._generate_get_presign_url(batch=file_list[i : i + BATCH_SIZE])
            )
        for item in file_list:
            local_path = Path(self._get_file_path(item.file_name))
            presigned_url = presigned_urls[self._get_obj_key(item.url)]
            path_args.append(local_path)
            url_args.append(presigned_url)

        with concurrent.futures.ThreadPoolExecutor(max_workers=BATCH_SIZE) as executor:
            logger.info("Starting multipart dataset file download")
            futures = [
                executor.submit(download_file_from_url, url, path)
                for url, path in zip(url_args, path_args)
            ]
            parts = [f.result() for f in concurrent.futures.as_completed(futures)]
            logger.info(f"Multipart dataset file download completed: {parts}")

    def _generate_get_presign_url(self, batch: List[DatasetFileInfo], expiration: int = 3600):
        resp = self._api_client.request(
            "POST",
            f"/api/v1/datasets/{self._dataset_id}/presign/getobject",
            json={
                "expiration": expiration,
                "obj_infos": [
                    {
                        constants.OBJ_KEY: self._get_obj_key(item.url),
                        constants.VERSION_ID: item.version_id,
                    }
                    for item in batch
                ],
            },
            raise_for_status=True,
        )
        return resp["data"]

    def _upload_files(self, file_list: List[DatasetFileInfo]):
        """
        upload file to the specific path

        Args:
            file_list (List[DatasetFileInfo]):
        Returns:
            List[DatasetFileInfo]
        """
        if len(file_list) == 0:
            logger.info("No files to upload")
            return file_list
        url_args = []
        path_args = []
        presigned_urls = {}
        for i in range(0, len(file_list), BATCH_SIZE):
            resp = self._api_client.request(
                "POST",
                f"/api/v1/datasets/{self._dataset_id}/presign/putobject",
                json={
                    "obj_keys": [self._get_obj_key(f.url) for f in file_list[i : i + BATCH_SIZE]]
                },
                raise_for_status=True,
            )
            presigned_urls.update(resp["data"])
        for item in file_list:
            url_args.append(presigned_urls[self._get_obj_key(item.url)])
            path_args.append(Path(self._get_file_path(item.file_name)))
        with concurrent.futures.ThreadPoolExecutor(max_workers=BATCH_SIZE) as executor:
            logger.info("Starting multipart dataset file upload")
            futures = [
                executor.submit(GantryDataset._multipart_upload_helper, url, path, item)
                for url, path, item in zip(url_args, path_args, file_list)
            ]
            parts = [f.result() for f in concurrent.futures.as_completed(futures)]
            logger.info(f"Multipart dataset file upload completed: {parts}")
        return file_list

    @staticmethod
    def _multipart_upload_helper(url: str, path: Path, item: DatasetFileInfo):
        version_id = upload_file_to_url(presigned_url=url, local_path=path)
        item.version_id = version_id
        return item.file_name

    def _generate_put_presign_url(self, batch: List[DatasetFileInfo]):
        resp = self._api_client.request(
            "POST",
            f"/api/v1/datasets/{self._dataset_id}/presign/putobject",
            json={"obj_keys": [self._get_obj_key(f.url) for f in batch]},
            raise_for_status=True,
        )
        return resp["data"]

    def _extract_feature_type(self) -> Dict[str, Any]:
        """
        Read feature type from dataset_config and convert them to huggingface feature type
        Returns:
            A key value pair of Dict where key is the column name and value is huggingface
            feature type
        """
        config = yaml.safe_load(
            (Path(self.workspace) / self.dataset_name / constants.DATASET_CONFIG_FILE).open(
                mode="r"
            )
        )
        return self._get_feature_map_from_config(config=config)

    def pull_huggingface_dataset_by_version(
        self,
        version_id: Optional[str] = None,
        hf_dataset_name: Optional[str] = None,
        streaming=False,
    ):
        """
        This function will pull the dataset tabular files from gantry server and convert it
        to huggingface dataset. This method will not pull data to your local dataset folder, it
        will only load dataset tabular data from a remote server into a huggingface dataset object.

        If there is no version_id specified, it will load the latest version of the dataset.
        If streaming set to True, it will return an `IterableDataset` or `IterableDatasetDict` instead
        of a Dataset or DatasetDict.

        Refer to `HuggingFace Documentation
        <https://huggingface.co/docs/datasets/stream>`_ for more details about streaming.

        Example usage:

        .. code-block:: python

            import gantry.dataset as gdataset

            dataset = gdataset.get_dataset(dataset_name)

            # download your dataset as huggingface dataset
            hf_dataset = dataset.pull_huggingface_dataset_by_version(version_id="version_id")

            # pull the dataset as IterableDataset with streaming
            hf_dataset = dataset.pull_huggingface_dataset_by_version(version_id="version_id", streaming=True) # noqa: E501

        Args:
            version_id (Optional[str], optional): dataset version id to load. Defaults to latest.
            hf_dataset_name (Optional[str], optional): param to overwrite the huggingface
                                                dataset name. Note this param must be a valid
                                                python identifier.
            streaming (bool, default ``False``): If set to True, don't download the data files.
                                                 Instead, it streams the data progressively while
                                                 iterating on the dataset. An IterableDataset or
                                                 IterableDatasetDict is returned instead
                                                 in this case.
        """
        if not hf_dataset_name:
            hf_dataset_name = self._clean_variable_name(self.dataset_name)

        if not GantryDataset._is_valid_python_identifier(hf_dataset_name=hf_dataset_name):
            return

        manifest_presign_url = self._get_manifest_presign_url(version_id=version_id)
        tabular_files = self._get_tabular_files(manifest_presign_url=manifest_presign_url)
        features = self._extract_feature_schema_from_s3(manifest_presign_url=manifest_presign_url)
        if not features:
            raise DataSchemaMismatchError()
        training_files = []
        expiration = 604800 if streaming else 7200  # 7 days if streaming, 2 hours otherwise
        for i in range(0, len(tabular_files), BATCH_SIZE):
            batch = tabular_files[i : i + BATCH_SIZE]
            presigned_urls = self._generate_get_presign_url(batch=batch, expiration=expiration)
            training_files.extend(presigned_urls.values())

        if not training_files:
            raise NoTabularDataError()

        env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(
                os.path.join(os.path.dirname(os.path.realpath(__file__)), "templates")
            )
        )

        template = env.get_template("load_script_template.py.jinja")

        content = template.render(
            training_file_list=training_files,
            dataset_version="1.0.0",  # Format should be x.y.z with {{x,y,z}} being digits
            dataset_name=hf_dataset_name,
            features=features,
        )
        hf_loading_script = (
            Path(self.workspace) / self.dataset_name / constants.HF_FOLDER / f"{hf_dataset_name}.py"
        )
        hf_loading_script.parent.mkdir(parents=True, exist_ok=True)
        with hf_loading_script.open("w") as f:
            f.write(content)

        try:
            hf_dataset = datasets.load_dataset(
                str(hf_loading_script), streaming=streaming, split="train"
            )
        except datasets.builder.DatasetGenerationError as e:
            raise DataSchemaMismatchError from e

        return hf_dataset

    def _get_manifest_presign_url(self, version_id: Optional[str] = None) -> str:
        target_commit = (
            self._get_commit(commit_id=version_id) if version_id else self._get_latest_commit()
        )

        metadata_version_id = target_commit[constants.METADATA_S3_FILE_VERSION]
        manifest_s3_url = self._get_s3_url(constants.DATASET_MANIFEST_FILE)

        return self._generate_get_presign_url(
            [
                DatasetFileInfo(
                    file_name=constants.DATASET_MANIFEST_FILE,
                    url=manifest_s3_url,
                    version_id=metadata_version_id,
                )
            ]
        )[self._get_obj_key(manifest_s3_url)]

    def _get_tabular_files(self, manifest_presign_url: str) -> List[DatasetFileInfo]:
        """
        Get tabular file list based on the version id

        Args:
            version_id (Optional[str], optional): dataset version. Defaults to None.

        Returns:
            List[DatasetFileInfo]: tabular file list
        """
        tabular_files = []
        for line in read_lines_from_url(manifest_presign_url):
            file_info = json.loads(line)
            if file_info[constants.FILE_NAME].startswith(constants.TABULAR_MANIFESTS) and file_info[
                constants.FILE_NAME
            ].endswith(constants.CSV_SUFFIX):
                tabular_files.append(
                    DatasetFileInfo(
                        file_name=file_info[constants.FILE_NAME],
                        url=file_info[constants.URL],
                        version_id=file_info[constants.VERSION_ID],
                    )
                )
        return tabular_files

    def _extract_feature_schema_from_s3(self, manifest_presign_url: str):
        for line in read_lines_from_url(manifest_presign_url):
            file_info = json.loads(line)
            if file_info[constants.FILE_NAME] == constants.DATASET_CONFIG_FILE:
                # read config file
                # generate the feature map
                config_presign_url = self._generate_get_presign_url(
                    [
                        DatasetFileInfo(
                            file_name=file_info[constants.FILE_NAME],
                            url=file_info[constants.URL],
                            version_id=file_info[constants.VERSION_ID],
                        )
                    ]
                )[self._get_obj_key(file_info[constants.URL])]

                # generate presign url for config file
                response = requests.get(config_presign_url)
                config = yaml.safe_load(response.text)

                return GantryDataset._get_feature_map_from_config(config=config)

    @staticmethod
    def _get_feature_map_from_config(config):
        columns = dict()
        if constants.DATASET_FEATURES_KEY in config:
            for feature_name, feature_dtype in config[constants.DATASET_FEATURES_KEY].items():
                columns[feature_name] = constants.GANTRY_2_HF_DTYPE[feature_dtype]

        if constants.DATASET_FEEDBACK_KEY in config:
            for feature_name, feature_dtype in config[constants.DATASET_FEEDBACK_KEY].items():
                columns[feature_name] = constants.GANTRY_2_HF_DTYPE[feature_dtype]

        return columns

    def get_huggingface_dataset(
        self, hf_dataset_name: Optional[str] = None, streaming: bool = False
    ):
        """
        This function will convert your local dataset folder to a huggingface dataset. Internally
        Gantry SDK will create a huggingface load script based on the dataset config and then load
        the dataset into huggingface pyarrow dataset.
        If streaming set to True, it will return
        an IterableDataset or IterableDatasetDict instead of a Dataset or DatasetDict.

        Refer to `HuggingFace documentation
        <https://huggingface.co/docs/datasets/stream>`_ for more
        details about streaming.

        Example usage:

        .. code-block:: python

            import gantry.dataset as gdataset

            dataset = gdataset.get_dataset(dataset_name)
            dataset.pull()

            # load the dataset into huggingface dataset
            hf_dataset = dataset.get_huggingface_dataset()

            # load the dataset into huggingface dataset with streaming
            hf_dataset = dataset.get_huggingface_dataset(streaming=True)

        Args:
            versoion_id (Optional[str], optional): dataset version id to load. Defaults to latest.
            hf_dataset_name (Optional[str], optional): param to overwrite the huggingface
                                                dataset name. Note this param must be a valid
                                                python identifier.
            streaming (bool, default ``False``): If set to True, don't download the data files.
                                                 Instead, it streams the data progressively while
                                                 iterating on the dataset. An IterableDataset or
                                                 IterableDatasetDict is returned instead
                                                 in this case.
        """

        if not hf_dataset_name:
            hf_dataset_name = GantryDataset._clean_variable_name(self.dataset_name)

        if not GantryDataset._is_valid_python_identifier(hf_dataset_name=hf_dataset_name):
            return

        tabular_manifests_path = (
            Path(self.workspace) / self.dataset_name / constants.TABULAR_MANIFESTS
        ).resolve()

        if not os.path.exists(tabular_manifests_path) or not any(
            Path(tabular_manifests_path).iterdir()
        ):
            raise NoTabularDataError()

        env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(
                os.path.join(os.path.dirname(os.path.realpath(__file__)), "templates")
            )
        )

        template = env.get_template("load_script_template.py.jinja")

        training_files = [str(p) for p in list_all_files(tabular_manifests_path.resolve())]
        features = self._extract_feature_type()

        # verify the csvs agree with the schema
        for csv_path in training_files:
            with open(csv_path) as data_csv:
                cols = next(csv.reader(data_csv))

                if set(cols) != set(features):
                    raise DataSchemaMismatchError()

        content = template.render(
            training_file_list=training_files,
            dataset_version="1.0.0",  # Format should be x.y.z with {{x,y,z}} being digits
            dataset_name=hf_dataset_name,
            features=features,
        )
        hf_loading_script = (
            Path(self.workspace) / self.dataset_name / constants.HF_FOLDER / f"{hf_dataset_name}.py"
        )
        os.makedirs(hf_loading_script.parent, exist_ok=True)
        with hf_loading_script.open("w") as f:
            f.write(content)

        try:
            hf_dataset = datasets.load_dataset(
                str(hf_loading_script), streaming=streaming, split="train"
            )
        except datasets.builder.DatasetGenerationError as e:
            raise DataSchemaMismatchError from e

        return hf_dataset

    @staticmethod
    def _is_valid_python_identifier(hf_dataset_name):
        if (
            (not hf_dataset_name)
            or (not hf_dataset_name.isidentifier())
            or iskeyword(hf_dataset_name)
        ):
            logger.error(
                "Can't get a valid hugging face dataset name, the name has to be a valid python"
                + "identifier. Please use hf_dataset_name param to set a valid name and try again."
            )
            return False
        else:
            return True

    def _prune_commit_info(self, commit_info):
        return {
            "version_id": commit_info["id"],
            "dataset": self.dataset_name,
            "message": commit_info["message"],
            # TODO:// uncomment the following line once we add sdk support for user version
            # "version_name": commit_info["user_version"],
            "created_at": commit_info["created_at"],
            "created_by": commit_info["created_by"],
            "is_latest_version": commit_info["is_latest_commit"],
        }

    @staticmethod
    def _clean_variable_name(ds_name):
        # Remove invalid characters
        ds_name = re.sub("[^0-9a-zA-Z_]", "", ds_name)
        # Remove leading characters until we find a letter or underscore
        ds_name = re.sub("^[^a-zA-Z_]+", "", ds_name)
        return ds_name



import json
from typing import Optional

import click
from click_aliases import ClickAliasedGroup

import gantry
from gantry.api_client import APIClient


@click.group(cls=ClickAliasedGroup)
def data_connector():
    """
    Use this to manage your data connectors with gantry.
    """
    pass


@data_connector.command(
    aliases=["create"],
    help="Creates a new data connector to your data source with Gantry.",
)
@click.option(
    "--name", type=click.STRING, required=True, help="A unique name of the data connector"
)
@click.option(
    "--connection-type",
    type=click.Choice(
        ["SNOWFLAKE", "S3"],
        case_sensitive=False,
    ),
    required=True,
    help="Type of the source data connection",
)
@click.option(
    "--database-name",
    type=click.STRING,
    required=False,
    default="",
    help="Name of the source database. WARNING: this field is DEPRECATED and will be removed in a "
    "future release. Please provide database name to 'options' field.",
)
@click.option(
    "--secret-name",
    type=click.STRING,
    required=True,
    help="Name of the secret registered with Gantry",
)
@click.option(
    "--description", type=click.STRING, required=True, help="Description of the data connector"
)
@click.option(
    "--options",
    type=click.STRING,
    required=True,
    help="""
JSON string of the options for the data connector.

For SNOWFLAKE connection type, the options should be:
    {
        "snowflake_database_name": "string",
        "snowflake_schema_name": "string",
        "snowflake_table_name": "string"
    }

For S3 connection type, the options should be:
    {
        "s3_bucket_name": "string",
        "s3_path_prefix": "string",
        "s3_filetype": "string",
    }
    S3 bucket name is the one provided in the secret you have registered with Gantry.
""",
)
def create(
    name: str,
    connection_type: str,
    database_name: str,
    secret_name: str,
    description: str,
    options: str,
):
    gantry.init()
    api_client: APIClient = gantry.get_client().log_store._api_client
    request_body = {
        "name": name,
        "connection_type": connection_type,
        "database_name": database_name,
        "secret_name": secret_name,
        "description": description,
        "options": json.loads(options),
    }

    resp = api_client.request(
        "POST", "/api/v1/data-connectors/sources", json=request_body, raise_for_status=True
    )

    click.secho("--> ", nl=False, fg="cyan")
    click.secho(f"A source data connector has been created.\n {json.dumps(resp['data'], indent=4)}")


@data_connector.command(
    aliases=["list"],
    help="List data connectors registered with Gantry.",
)
def list():
    gantry.init()
    api_client: APIClient = gantry.get_client().log_store._api_client
    resp = api_client.request("GET", "/api/v1/data-connectors", raise_for_status=True)

    click.secho(f"--> A list of registered data connectors:\n {json.dumps(resp['data'], indent=4)}")


@data_connector.command(
    aliases=["delete"],
    help="Deletes a data connector.",
)
@click.option(
    "--name", type=click.STRING, required=True, help="A unique name of the data connector"
)
def delete(name: str):
    gantry.init()
    api_client: APIClient = gantry.get_client().log_store._api_client
    resp = api_client.request(
        "DELETE", f"/api/v1/data-connectors/sources/{name}", raise_for_status=True
    )

    click.secho(
        f"--> Deleted the source data connector {name}:\n {json.dumps(resp['data'], indent=4)}"
    )


@data_connector.command(
    aliases=["list_pipelines"],
    help="Lists the pipelines for a given app name and status.",
)
@click.option("--app-name", type=click.STRING, required=True, help="The name of the application")
@click.option(
    "--status",
    required=False,
    type=click.Choice(
        ["enabled", "disabled", "any"],
        case_sensitive=False,
    ),
    help="The status of the data connectors",
)
def list_pipelines(app_name: str, status: Optional[str]):
    gantry.init()
    api_client: APIClient = gantry.get_client().log_store._api_client
    enabled_status = "True" if status == "enabled" else "False" if status == "disabled" else "True"
    resp = api_client.request(
        "GET",
        f"/api/v1/data-connectors/pipelines?app_name={app_name}&enabled_status={enabled_status}",
        raise_for_status=True,
    )

    click.secho(
        f"--> The following pipelines are given for \"{app_name}\" app with status \"{status}\":\n\
{json.dumps(resp['data'], indent=4)}"
    )


@data_connector.command(
    aliases=["list_pipeline_operations"],
    help="Lists pipeline operations for a given pipeline name, appname, and status.",
)
@click.option("--pipeline-name", type=click.STRING, required=True, help="The data pipeline name")
@click.option("--app-name", type=click.STRING, required=True, help="The name of the application")
@click.option(
    "--status",
    required=False,
    type=click.Choice(
        ["QUEUED", "RUNNING", "SUCCEEDED", "FAILED"],
        case_sensitive=False,
    ),
    help="The pipeline operation status to filter by",
)
def list_pipeline_operations(pipeline_name: str, app_name: str, status: Optional[str] = ""):
    gantry.init()
    api_client: APIClient = gantry.get_client().log_store._api_client
    query = (
        f"/api/v1/data-connectors/pipeline-operations?app_name={app_name}"
        f"&pipeline_name={pipeline_name}&status={status or ''}"
    )
    resp = api_client.request(
        "GET",
        query,
        raise_for_status=True,
    )

    echo = (
        f"--> The following pipeline operations are given for APP={app_name} "
        f"and PIPELINE={pipeline_name} and STATUS={status} : )\n "
        f"{json.dumps(resp['data'], indent=4)}"
    )
    click.secho(echo)


@data_connector.command(
    aliases=["delete"],
    help="Deletes a data pipeline.",
)
@click.option("--pipeline-name", type=click.STRING, required=True, help="The data pipeline name")
@click.option("--app-name", type=click.STRING, required=True, help="The name of the application")
@click.option("--app-version", type=click.STRING, required=False, help="The version of the model")
def delete_pipeline(pipeline_name: str, app_name: str, app_version: str):
    gantry.init()
    api_client: APIClient = gantry.get_client().log_store._api_client
    query = (
        f"/api/v1/data-connectors/pipelines?pipeline_name={pipeline_name}&app_name={app_name}"
        f"&app_version={app_version}"
    )
    resp = api_client.request(
        "DELETE",
        query,
        raise_for_status=True,
    )

    click.secho(f"--> Deleted the pipeline{pipeline_name}:\n {json.dumps(resp['data'], indent=4)}")


if __name__ == "__main__":
    data_connector()



import datetime
import uuid
from typing import List, Optional

import dateparser
from pydantic import BaseModel, validator

from gantry.automations.curators.selectors import Selector


class CreateCuratorRequest(BaseModel):
    application_name: str
    name: str
    start_on: datetime.datetime
    curation_interval: datetime.timedelta
    curated_dataset_name: Optional[str]
    curation_delay: datetime.timedelta = datetime.timedelta(days=0)
    curate_past_intervals: bool = False
    selectors: List[Selector]


class UpdateCuratorRequest(BaseModel):
    name: str
    new_name: Optional[str]
    new_curated_dataset_name: Optional[str]
    curation_interval: Optional[datetime.timedelta]
    selectors: Optional[List[Selector]]
    allow_create_new_dataset: bool = False


class EnableCuratorRequest(BaseModel):
    name: str
    enable: bool


class CuratorInfo(BaseModel):
    id: uuid.UUID
    application_name: str
    name: str
    curated_dataset_name: str
    start_on: datetime.datetime
    curation_interval: datetime.timedelta
    curation_delay: datetime.timedelta
    curate_past_intervals: bool
    created_at: datetime.datetime
    selectors: List[Selector]

    @validator("start_on", "created_at", pre=True)
    def parse_datetime(cls, v):
        return dateparser.parse(v)


class DeletedCuratorInfo(BaseModel):
    curated_dataset_name: str
    id: uuid.UUID
    application_name: str
    deleted_at: datetime.datetime



import datetime

import mock
import numpy as np
import pandas as pd
import pytest
import responses

from gantry.api_client import APIClient
from gantry.applications.client import ApplicationClient
from gantry.applications.core import Application
from gantry.query.core.dataframe import GantryDataFrame
from gantry.query.time_window import TimeWindow

HOST = "https://test-api"
CURRENT_TIME = datetime.datetime.utcnow()


@pytest.fixture
def test_api_client():
    return APIClient(origin=HOST)


@pytest.fixture
def test_client(test_api_client):
    return ApplicationClient(api_client=test_api_client)


@pytest.mark.parametrize(
    "test_timestamps",
    [None, [CURRENT_TIME] * 2],
)
@pytest.mark.parametrize(
    "test_feedbacks",
    [
        [{"A": 200}, {"A": 201}],
        pd.DataFrame.from_dict({"A": [200, 201]}),
        pd.Series(name="A", data=[200, 201]),
        np.array([{"A": 200}, {"A": 201}]),
    ],
)
@pytest.mark.parametrize(
    "test_outputs",
    [
        [{"B": 300}, {"B": 301}],
        pd.DataFrame.from_dict({"B": [300, 301]}),
        pd.Series(name="B", data=[300, 301]),
        np.array([{"B": 300}, {"B": 301}]),
    ],
)
@pytest.mark.parametrize(
    "test_inputs",
    [
        [{"A": 100}, {"A": 101}],
        pd.DataFrame.from_dict({"A": [100, 101]}),
        pd.Series(name="A", data=[100, 101]),
        np.array([{"A": 100}, {"A": 101}]),
    ],
)
@mock.patch("gantry.logger.client.Gantry.log")
def test_log_with_run(
    mock_log,
    test_inputs,
    test_outputs,
    test_feedbacks,
    test_client,
    test_api_client,
    test_timestamps,
):
    application = Application(name="test_application_name", api_client=test_api_client)
    assert application._name == "test_application_name"
    m = mock.Mock()
    with mock.patch("gantry.logger.main._CLIENT", m):
        getattr(m, "log").return_value = "return_value"
        with application.start_run(name="test-run", tags={"run": "test"}) as run:
            result = application.log(
                inputs=test_inputs,
                outputs=test_outputs,
                feedbacks=test_feedbacks,
                timestamps=CURRENT_TIME,
                join_keys=["54321", "67890"],
            )
            assert result == "return_value"
            assert run.name == "test-run"
            assert run.tags == {"run": "test"}
            assert run.join_keys == []
            getattr(m, "log").assert_called_once_with(
                application="test_application_name",
                inputs=test_inputs,
                outputs=test_outputs,
                feedbacks=test_feedbacks,
                ignore_inputs=None,
                timestamps=CURRENT_TIME,
                sample_rate=1.0,
                sort_on_timestamp=True,
                tags=None,
                row_tags=None,
                global_tags=None,
                join_keys=["54321", "67890"],
                as_batch=False,
                run_tags={"run": "test"},
                run_id=run.run_id,
                version=None,
            )


def test_create_dataset(test_api_client):
    application = Application(name="test_application_name", api_client=test_api_client)
    m = mock.Mock()
    with mock.patch("gantry.dataset.main._DATASET", m):
        getattr(m, "create_dataset").return_value = "return_value"
        result = application.create_dataset(
            name="test_dataset_name", bucket_name="test-dataset-bucket"
        )
        assert result == "return_value"
        getattr(m, "create_dataset").assert_called_once_with(
            name="test_dataset_name",
            bucket_name="test-dataset-bucket",
            app_name="test_application_name",
        )


def test_list_datasets(test_api_client):
    application = Application(name="test_application_name", api_client=test_api_client)
    m = mock.Mock()
    with mock.patch("gantry.dataset.main._DATASET", m):
        getattr(m, "list_datasets").return_value = "return_value"
        result = application.list_datasets(include_deleted=True)
        assert result == "return_value"
        getattr(m, "list_datasets").assert_called_once_with(
            include_deleted=True,
            model_node_id=application._id,
        )


def test_get_schema(test_api_client):
    application = Application(name="test_application_name", api_client=test_api_client)
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/applications/{application._name}/schemas",
            json={
                "response": "ok",
                "data": {"schema": "return_value"},
            },
            headers={"Content-Type": "application/json"},
        )
        schema = application.get_schema()
        assert schema == {"schema": "return_value"}


def test_list_workspaces(test_api_client):
    application = Application(name="test_application_name", api_client=test_api_client)
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/workspaces",
            json={
                "response": "ok",
                "data": {"workspace": "return_value"},
            },
            headers={"Content-Type": "application/json"},
        )
        workspaces = application.list_workspaces()
        assert workspaces == {"workspace": "return_value"}


def test_list_automations(test_api_client):
    application = Application(name="test_application_name", api_client=test_api_client)
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/automations",
            json={
                "response": "ok",
                "data": {"automations": "return_value"},
            },
            headers={"Content-Type": "application/json"},
        )
        automations = application.list_automations()
        assert automations == {"automations": "return_value"}


def test_get_run(test_api_client):
    application = Application(name="test_application_name", api_client=test_api_client)
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/applications/{application._id}/jobs",
            json={
                "response": "ok",
                "data": {"jobs": "return_value"},
            },
            headers={"Content-Type": "application/json"},
        )
        run = application.get_run(name="test_run")
        assert run == "return_value"

    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/applications/{application._id}/jobs",
            json={
                "response": "ok",
                "data": {"jobs": "return_value"},
            },
            headers={"Content-Type": "application/json"},
        )
        run = application.get_run(tags={"run": "test"})
        assert run == "return_value"


def test_query(test_api_client):
    application = Application(name="test_application_name", api_client=test_api_client)
    assert application._name == "test_application_name"

    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/models/{application._name}/schemas",
            json={
                "response": "ok",
                "data": {"id": "return_value"},
            },
            headers={"Content-Type": "application/json"},
        )
        expected_query = GantryDataFrame(
            api_client=test_api_client,
            application=application._name,
            start_time=CURRENT_TIME - datetime.timedelta(days=1),
            end_time=CURRENT_TIME,
            version=None,
            env="test",
            filters=[{"feature_name": "inputs.A", "lower_bound": 0, "upper_bound": 10}],
            tags={"query": "test"},
        )
        actual_query = application.query(
            time_window=TimeWindow(
                start_time=CURRENT_TIME - datetime.timedelta(days=1), end_time=CURRENT_TIME
            ),
            version=None,
            env="test",
            filters=[{"feature_name": "inputs.A", "lower_bound": 0, "upper_bound": 10}],
            tags={"query": "test"},
        )
        assert isinstance(actual_query, GantryDataFrame)
        assert actual_query.start_time == expected_query.start_time
        assert actual_query.end_time == expected_query.end_time
        assert actual_query.version == expected_query.version
        assert actual_query.env == expected_query.env
        assert actual_query.filters == expected_query.filters
        assert actual_query.tags == expected_query.tags


def test_save_query(test_api_client):
    application = Application(name="test_application_name", api_client=test_api_client)
    assert application._name == "test_application_name"

    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/models/{application._name}/schemas",
            json={
                "response": "ok",
                "data": {"id": str(application._id)},
            },
            headers={"Content-Type": "application/json"},
        )
        resp.add(
            resp.POST,
            f"{HOST}/api/v1/queries",
            json={
                "response": "ok",
                "data": {"query": "test"},
            },
            headers={"Content-Type": "application/json"},
        )
        resp.add(
            resp.PATCH,
            f"{HOST}/api/v1/queries",
            json={
                "response": "ok",
                "data": {"query": "test"},
            },
            headers={"Content-Type": "application/json"},
        )
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/applications/{application._id}/queries",
            json={
                "response": "ok",
                "data": [],
            },
            headers={"Content-Type": "application/json"},
        )
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/applications/{application._id}/queries/test_query",
            json={
                "response": "ok",
                "data": {
                    "start_time": (CURRENT_TIME - datetime.timedelta(days=1)).isoformat(),
                    "end_time": CURRENT_TIME.isoformat(),
                    "filters": [{"feature_name": "inputs.A", "lower_bound": 0, "upper_bound": 10}],
                    "tags": [{"tag_name": "query", "tag_value": "test"}],
                },
            },
            headers={"Content-Type": "application/json"},
        )
        actual_query = application.query(
            time_window=TimeWindow(
                start_time=CURRENT_TIME - datetime.timedelta(days=1),
                end_time=CURRENT_TIME,
            ),
            version=None,
            env=None,
            filters=[{"feature_name": "inputs.A", "lower_bound": 0, "upper_bound": 10}],
            tags={"query": "test"},
        )
        application.save_query(name="test_query", query=actual_query)
        saved_query = application.get_query(name="test_query")
        # Timezone difference, but the assert is still correct
        # assert saved_query.start_time == actual_query.start_time
        # assert saved_query.end_time == actual_query.end_time
        assert saved_query.version == actual_query.version
        assert saved_query.env == actual_query.env
        assert saved_query.filters == actual_query.filters
        assert saved_query.tags == actual_query.tags

        # Add a mock response that there is a query with the same name in the list
        # to test if it updates the query instead of saving.
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/applications/{application._id}/queries",
            json={
                "response": "ok",
                "data": [{"name": "test_query", "query": "test"}],
            },
            headers={"Content-Type": "application/json"},
        )
        actual_query = application.query(
            time_window=TimeWindow(
                start_time=CURRENT_TIME - datetime.timedelta(days=1), end_time=CURRENT_TIME
            ),
            version=None,
            env="test",
            filters=[{"feature_name": "inputs.A", "lower_bound": 0, "upper_bound": 5}],
            tags={"query": "test"},
        )
        application.save_query(name="test_query", query=actual_query)
        application.get_query(name="test_query")



from enum import Enum
from typing import List, Optional, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal  # type: ignore

from pydantic import BaseModel, root_validator, validator

# Database constraint
MAX_NAME_LEN = 255
TIME_COLUMN_NAME = "__time"


class OrderingDirection(str, Enum):
    """
    Direction specifiying how to order a sort.
    """

    ASCENDING = "ascending"
    DESCENDING = "descending"


class SamplingMethod(str, Enum):
    """
    Sampling method to use when selecting a subset of data.

    Selected data will be sampled using one of the methods enumerated here.
    """

    uniform = "uniform"
    ordered = "ordered"
    stratified = "stratified"


class StratificationMode(str, Enum):
    """
    Stratification mode to use when selecting a subset of data. Currently,
    stratification is only supported for categorical data.

    Selected data will be stratified using one of the modes enumerated here.

    In "strict" mode, the limit of records per curation is respected, but the proportion of records
    in each stratum may be violated, unlike the ProportionalStratificationCurator, which will
    respect the proportion at the (potential) cost of violating the limit per curation.

    Additionally, it is worth noting that the number of records to be added to each stratum
    is calulated exactly, leading to poentially fractional assignments. The integer part
    of the assignment is taken, and the total remainder is added, up to the limit, randomly.

    Example:
        Let's say there are 100 records: 51 records with the value "A" for the field, 15
        records with the value "B" for the field, and 16 records with the value "C" for the
        field, 17 records with the value "D" for the field, and 1 record with the value "E".

        If the limit is 10, then the curator will return 10 records, with 5 records with the
        value "A" for the field, 1 record with "B" for the field, 1 record with "C" for the
        field, 1 record with "D" for the field, 0 records with "E" for the field, and the
        remaining 2 records will be randomly selected from the remaining records.

    In "proportional" mode, the proportion of records in each stratum is respected,
    but the limit per curation may be violated, unlike the StrictStratificationCurator,
    which will respect the limit at the (potential) cost of violating the proportion of
    records in each stratum.

    Additionally, it is worth noting that proportions are rounded to even numbers, so
    1.5 will be rounded to 2, and 0.5 will be rounded to 0.

    Example:
        Let's say there are 100 records: 51 records with the value "A" for the field, 15
        records with the value "B" for the field, and 16 records with the value "C" for the
        field, 17 records with the value "D" for the field, and 1 record with the value "E".

        If the limit is 10, then the curator will return 11 records, with 5 records with the
        value "A" for the field, 2 records with the value "B" for the field, 2 records with
        the value "C" for the field, 2 records with the value "D" for the field, and 0 records
        with the value "E".

    In "balanced" mode, the number of records per stratum is balanced with respect to the
    least represented stratum.
    """

    strict = "strict"
    proportional = "proportional"
    balanced = "balanced"


class TagFilter(BaseModel):
    """Filter for tags"""

    name: str
    value: str


class Filter(BaseModel):
    """Base class for filters. All filters must have a field."""

    field: str


class BoundsFilter(Filter):
    """
    Filter for bounds. Must have either an upper or lower bound; both
    can be specified. Default bounds are exclusive, but inclusive bounds
    can be given instead.
    """

    upper: Optional[Union[float, int]] = None
    inclusive_upper: Optional[Union[float, int]] = None
    lower: Optional[Union[float, int]] = None
    inclusive_lower: Optional[Union[float, int]] = None

    @root_validator
    def validate_bounds(cls, values):
        if (
            values["upper"] is None
            and values["lower"] is None
            and values["inclusive_upper"] is None
            and values["inclusive_lower"] is None
        ):
            raise ValueError("Must have either an upper or lower bound")
        if values["upper"] is not None and values["inclusive_upper"] is not None:
            raise ValueError("Cannot have inclusive and exclusive upper bound")
        if values["lower"] is not None and values["inclusive_lower"] is not None:
            raise ValueError("Cannot have inclusive and exclusive lower bound")

        return values


class EqualsFilter(Filter):
    """
    Filter for equality. Must have an equals field.

    The equals field can be a boolean, string, float, or int.
    """

    equals: Union[bool, float, int, str]


class ContainsFilter(Filter):
    """
    Filter for string containment. The specified field
    must be a string.
    """

    contains: str


ALL_FILTERS_T = Union[BoundsFilter, EqualsFilter, ContainsFilter]


class Sampler(BaseModel):
    """Base class for samplers."""

    pass


class UniformSampler(Sampler):
    """
    Sampler for uniform sampling.

    This sampler will select a subset of the data uniformly at random.
    """

    sample: Literal[SamplingMethod.uniform] = SamplingMethod.uniform


class OrderedSampler(Sampler):
    """
    Sampler for ordered sampling.

    This sampler will select a subset of the data ordered by the specified field.
    Fields can be ordered in ascending or descending order.
    """

    sample: Literal[SamplingMethod.ordered] = SamplingMethod.ordered
    field: str
    sort: Literal[OrderingDirection.ASCENDING, OrderingDirection.DESCENDING]


class StratifiedSampler(Sampler):
    """
    Sampler for stratified sampling.

    This sampler will select a subset of the data stratified by the specified field.
    Stratification is only supported for categorical data. The mode specifies how to
    stratify the data. See the documentation for StratificationMode for more details.
    """

    field: str
    sample: Literal[SamplingMethod.stratified] = SamplingMethod.stratified
    mode: Literal[
        StratificationMode.strict, StratificationMode.proportional, StratificationMode.balanced
    ] = StratificationMode.proportional


ALL_METHODS_T = Union[OrderedSampler, UniformSampler, StratifiedSampler]


class Selector(BaseModel):
    """
    Base class for selectors. All selectors must have a method and a limit. The
    method specifies how to select a subset of the data, and the limit specifies
    the number of records to select. The filters field specifies any filters to
    apply to the data before selecting a subset.
    """

    method: ALL_METHODS_T = OrderedSampler(
        sample=SamplingMethod.ordered,
        field=TIME_COLUMN_NAME,
        sort=OrderingDirection.DESCENDING,
    )
    limit: int
    filters: List[ALL_FILTERS_T] = []
    tags: List[TagFilter] = []

    @validator("limit")
    def validate_limit(cls, value):
        if value <= 0:
            raise ValueError("Limit must be strictly positive")
        if value > 10_000:
            raise ValueError(
                "Currently, curation limit must not exceed 10,000 records per interval"
            )

        return value



# api calls like like "list all curators" live on the curator client
_CURATOR_CLIENT = None

# instantions of curators don't need all of the curator client's methods
# but also need to be able to make api calls (and we don't want to initialize
# a new client for each curator)
_API_CLIENT = None



import mock
import pytest

from gantry.dataset import main
from gantry.exceptions import ClientNotInitialized


@pytest.mark.parametrize(
    ["method", "params"],
    [
        ("create_dataset", {"name": "dataset_name"}),
        ("get_dataset", {"name": "dataset_name"}),
        ("list_dataset_versions", {"name": "dataset_name"}),
        ("list_datasets", {}),
        ("delete_dataset", {"name": "dataset_name"}),
    ],
)
def test_uninit_client_main(method, params):
    """Test all public methods from dataset module
    fail if client is not initialized
    """
    with mock.patch("gantry.dataset.main._DATASET", None):
        with pytest.raises(ClientNotInitialized):
            getattr(main, method)(**params)


@pytest.mark.parametrize("backend", ["", "/some/file", "tcp://other/protocol"])
def test_global_init_error(backend):
    with mock.patch.dict("os.environ", {"GANTRY_LOGS_LOCATION": backend}):
        with pytest.raises(ValueError):
            main._init()


def test_no_api_key_provided():
    with mock.patch.dict("os.environ", {"GANTRY_API_KEY": ""}):
        with pytest.raises(ValueError):
            main._init(api_key=None)


def test_init():
    """Test init"""
    passed_api_key = "some_key"
    with mock.patch.dict(
        "os.environ",
        {
            "GANTRY_API_KEY": passed_api_key,
            "GANTRY_LOGS_LOCATION": "https://foo/bar",
            "GANTRY_DATASET_WORKING_DIR": "working_dir",
        },
    ):
        main._init()
        assert main._DATASET._api_client._host == "https://foo/bar"
        assert main._DATASET._api_client._api_key == passed_api_key
        assert main._DATASET.working_directory == "working_dir"


def test_init_default():
    """Test init client with default values"""
    main._init(api_key="ABCD1234")
    assert main._DATASET._api_client._host == "https://app.gantry.io"
    assert main._DATASET._api_client._api_key == "ABCD1234"


@pytest.mark.parametrize(
    ["method", "params"],
    [
        ("create_dataset", {"name": "dataset_name"}),
        ("get_dataset", {"name": "dataset_name"}),
        ("list_dataset_versions", {"name": "dataset_name"}),
        ("list_datasets", {}),
        ("delete_dataset", {"name": "dataset_name"}),
        ("set_working_directory", {"working_dir": "path"}),
    ],
)
def test_dataset_methods(method, params):
    """Test all public methods from gantry module or gantry.main module
    resolve in the global _DATASET methods
    """
    m = mock.Mock()
    with mock.patch("gantry.dataset.main._DATASET", m):
        getattr(main, method)(**params)
        getattr(m, method).assert_called_once_with(**params)



import functools
import inspect
import logging
import os
from distutils.util import strtobool
from typing import Optional, Tuple, cast

import colorama
from typeguard import typechecked
from typing_extensions import Literal

from gantry.const import DEFAULT_ENV, PROD_API_URL
from gantry.exceptions import ClientNotInitialized
from gantry.logger.client import Gantry
from gantry.logger.stores import APILogStore

colorama.init(autoreset=True)
logger_obj = logging.getLogger(__name__)
logger_obj.addHandler(logging.NullHandler())

_CLIENT: Optional[Gantry] = None


def _client_alias(f):
    doc = "Alias for :meth:`gantry.logger.client.Gantry.{}`".format(f.__name__)
    signature = inspect.signature(getattr(Gantry, f.__name__))
    orig_doc = inspect.getdoc(getattr(Gantry, f.__name__))

    if orig_doc:
        doc += "\n\n{}".format(orig_doc)
    signature = signature.replace(parameters=tuple(signature.parameters.values())[1:])

    f.__doc__ = doc
    f.__signature__ = signature

    # This decorator also checks that the _CLIENT
    # has been initialized
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        if _CLIENT is None:
            raise ClientNotInitialized()

        return f(*args, **kwargs)

    return wrapper


LOGGING_LEVEL_T = Literal["DEBUG", "INFO", "WARNING", "CRITICAL", "ERROR"]


@typechecked
def _resolve_params(
    api_key: Optional[str] = None,
    logging_level: Optional[LOGGING_LEVEL_T] = None,
    environment: Optional[str] = None,
    send_in_background: Optional[bool] = None,
) -> Tuple[str, str, LOGGING_LEVEL_T, str, bool, bool]:
    api_key = api_key or os.environ.get("GANTRY_API_KEY")
    if api_key is None:
        raise ValueError("logger was not initialized with API Key")

    logs_location = cast(str, os.environ.get("GANTRY_LOGS_LOCATION", PROD_API_URL))
    logging_level = cast(
        LOGGING_LEVEL_T, logging_level or os.environ.get("GANTRY_LOGGING_LEVEL", "INFO")
    )
    environment = cast(str, environment or os.environ.get("GANTRY_ENVIRONMENT", DEFAULT_ENV))

    # TODO: use urlparse to grab the scheme instead
    if not (logs_location.startswith("http://") or logs_location.startswith("https://")):
        logger_obj.error("Currently only HTTP logs location supported")
        raise ValueError(
            "logger was not initialized with http endpoint. Only HTTP logs locations supported."
        )

    bypass_firehose = bool(strtobool(os.environ.get("GANTRY_BYPASS_FIREHOSE", "false")))

    if send_in_background is None:
        if "GANTRY_SEND_IN_BACKGROUND" in os.environ:
            send_in_background = bool(strtobool(os.environ["GANTRY_SEND_IN_BACKGROUND"]))
        else:
            # Set sync behavior in case app is running in a lambda (scenario in which
            # we can't start a thread).
            # https://docs.aws.amazon.com/lambda/latest/dg/configuration-envvars.html
            send_in_background = os.environ.get("AWS_LAMBDA_FUNCTION_NAME") is None

    return api_key, logs_location, logging_level, environment, bypass_firehose, send_in_background


def _init(
    api_key: Optional[str] = None,
    logging_level: Optional[LOGGING_LEVEL_T] = None,
    environment: Optional[str] = None,
    send_in_background: Optional[bool] = None,
):
    """
    Initialize the logger. Initialization should happen before submitting any data to Gantry.

    Args:
        api_key (optional, str): The Gantry API Key. You can also set this parameter by setting
            the env variable GANTRY_API_KEY.
        logging_level (optional, str): Set logging level for Gantry system logging.
            You can also set this parameter by setting the env variable GANTRY_LOGGING_LEVEL.
            Options are: DEBUG, INFO, WARNING, CRITICAL or ERROR.
            If not specified, it defaults to INFO.
        environment (optional, str): Set the value for the environment label attached
            to data instrumented. You can also set this parameter by setting the env
            variable GANTRY_ENVIRONMENT. If not provided, it defaults to "dev".
            The environment is essentially a tag attached to data. To override
            this value later, you can set an 'env' tag in the tags parameters
            in the ingestion methods.
        send_in_background (optional, bool): set whether Gantry logging methods should
            run synchronously or not. You can also set this value by setting
            the env variable GANTRY_SEND_IN_BACKGROUND.
            If not provided, it defaults to True unless running in an AWS lambda.
    """
    (
        api_key,
        logs_location,
        logging_level,
        environment,
        bypass_firehose,
        send_in_background,
    ) = _resolve_params(api_key, logging_level, environment, send_in_background)

    global _CLIENT

    log_store = APILogStore(
        location=logs_location,
        api_key=api_key,
        bypass_firehose=bypass_firehose,
        send_in_background=send_in_background,
    )
    _CLIENT = Gantry(log_store, environment, logging_level)

    logger_obj.debug("Gantry logger initialized")

    if not _CLIENT.ping():
        logger_obj.warning(
            "Gantry services not reachable. Check provided URL "
            f"[{logs_location}] is pointing to the correct address"
        )
        return

    if not _CLIENT.ready():
        logger_obj.warning("Gantry services won't receive traffic. Check if API Key is valid")


@_client_alias
def ping():
    return _CLIENT.ping()


@_client_alias
def ready():
    return _CLIENT.ready()


@_client_alias
def log_file(*args, **kwargs):
    return _CLIENT.log_file(*args, **kwargs)  # type: ignore[union-attr]


@_client_alias
def log_from_data_connector(*args, **kwargs):
    return _CLIENT.log_from_data_connector(*args, **kwargs)  # type: ignore[union-attr]


@_client_alias
def log_record(*args, **kwargs):
    return _CLIENT.log_record(*args, **kwargs)  # type: ignore[union-attr]


@_client_alias
def log_records(*args, **kwargs):
    return _CLIENT.log_records(*args, **kwargs)  # type: ignore[union-attr]


def get_client():
    return _CLIENT


def setup_logger(level: str = "INFO"):
    return Gantry.setup_logger(level)


@_client_alias
def log(*args, **kwargs):
    return _CLIENT.log(*args, **kwargs)  # type: ignore[union-attr]


@_client_alias
def generate_records(*args, **kwargs):
    return _CLIENT.generate_records(*args, **kwargs)  # type: ignore[union-attr]


@_client_alias
def _upload_data_as_batch(*args, **kwargs):
    return _CLIENT._upload_data_as_batch(*args, **kwargs)  # type: ignore[union-attr]



import datetime
from abc import ABC
from typing import Dict, List, Optional

from gantry.alerts import create_alert
from gantry.alerts.client import AlertsAggregation, AlertsCheck
from gantry.query.time_window import RelativeTimeWindow
from gantry.utils import from_isoformat_duration, to_datetime, to_isoformat_duration


class Trigger(ABC):
    def __init__(self):
        self._application_name = None

    def add_to_application(self, application_name: str):
        self._application_name = application_name

    def remove_from_application(self):
        self._application_name = None

    def to_dict(self):
        pass


class IntervalTrigger(Trigger):
    def __init__(
        self,
        start_on: datetime.datetime,
        interval: datetime.timedelta,
        delay: datetime.timedelta = datetime.timedelta(),
    ):
        super().__init__()
        self.start_on = start_on
        self.interval = interval
        self.delay = delay or datetime.timedelta()

    def is_triggered(self):
        raise NotImplementedError

    def to_dict(self) -> Dict:
        return {
            "type": "IntervalTrigger",
            "content": {
                "start_on": self.start_on.isoformat(),
                "interval": to_isoformat_duration(self.interval),
                "delay": to_isoformat_duration(self.delay),
            },
        }

    @classmethod
    def from_dict(cls, d) -> "IntervalTrigger":
        return cls(
            start_on=to_datetime(d["start_on"]),
            interval=from_isoformat_duration(d["interval"]),
            delay=from_isoformat_duration(d["delay"]),
        )


class AggregationTrigger(Trigger):
    def __init__(
        self,
        name: str,
        aggregation: AlertsAggregation,
        fields: List[str],
        lower_bound: float,
        upper_bound: float,
        evaluation_window: RelativeTimeWindow,
        tags: Optional[Dict] = None,
    ):
        super().__init__()
        self.name = name
        self.aggregation = aggregation
        self.fields = fields
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.evaluation_window = evaluation_window.window_length
        self.delay = evaluation_window.offset
        self.tags = tags

    def generate_alert(self) -> str:
        alert_check = AlertsCheck(
            self.aggregation, self.fields, self.lower_bound, self.upper_bound, "range_check"
        )
        evaluation_window = to_isoformat_duration(self.evaluation_window)
        delay = to_isoformat_duration(datetime.timedelta())
        if self.delay:
            delay = to_isoformat_duration(self.delay)
        alert_id = create_alert(
            self._application_name,
            self.name,
            [alert_check],
            evaluation_window,
            delay,
            self.tags,
        )
        return alert_id

    def to_dict(self) -> Dict:
        return {
            "type": "AggregationTrigger",
            "content": {
                "name": self.name,
                "aggregation": self.aggregation,
                "fields": self.fields,
                "lower_bound": self.lower_bound,
                "upper_bound": self.upper_bound,
                "evaluation_window": to_isoformat_duration(self.evaluation_window),
                "delay": to_isoformat_duration(self.delay)
                if self.delay
                else to_isoformat_duration(datetime.timedelta()),
                "tags": self.tags,
            },
        }

    @classmethod
    def from_dict(cls, d) -> "AggregationTrigger":
        return cls(
            d["name"],
            d["aggregation"],
            d["fields"],
            d["lower_bound"],
            d["upper_bound"],
            RelativeTimeWindow(
                window_length=from_isoformat_duration(d["evaluation_window"]),
                offset=from_isoformat_duration(d["delay"]),
            ),
            d["tags"],
        )



import datetime
import json
import logging
import uuid
from typing import List, Optional

from typeguard import typechecked

from gantry.api_client import APIClient
from gantry.automations.curators import globals
from gantry.automations.curators.models import (
    CreateCuratorRequest,
    CuratorInfo,
    EnableCuratorRequest,
    UpdateCuratorRequest,
)
from gantry.automations.curators.selectors import Selector
from gantry.dataset.gantry_dataset import GantryDataset
from gantry.dataset.main import get_dataset
from gantry.exceptions import (
    ClientNotInitialized,
    CuratorNotFoundError,
    GantryRequestException,
)

logger = logging.getLogger(__name__)


class CuratorClient:
    def __init__(self, api_client: APIClient) -> None:
        self._api_client = api_client

    @typechecked
    def list_curators(self, application_name: Optional[str] = None) -> List[str]:
        """Returns a list of the names of all the curators in the
        organization if no ``application_name`` is passed, and
        all the curators associated with the provided application
        if a name is passed.

        Args:
            application_name (str, optional): defaults to ``None``,
                application must exist.

        Returns:
            ``List[str]``: A list of the curators either in the organization or
            for a specific application.
        """
        return [
            curator_info.name
            for curator_info in self._get_all_curator_infos(application_name=application_name)
        ]

    @typechecked
    def get_all_curators(self, application_name: Optional[str] = None) -> List["Curator"]:
        """Returns a list of the curators in the organization if no
        ``application_name`` is passed, and all the curators associated
        with the provided application if a name is passed.

        Args:
            application_name (str, optional): defaults to ``None``,
                application must exist.

        Returns:
            ``List[Curator]``: A list of the curators either in the
            organization or for a specific application.
        """
        res = self._get_all_curator_infos(application_name=application_name)
        return [Curator.from_curator_info(self._api_client, curator_info) for curator_info in res]

    @typechecked
    def get_curator(self, name: str) -> "Curator":
        """
        Get a curator by name. If the curator does not exist, an exception will be raised.

        Args:
            name (str): the name of the curator to be retrieved.

        Returns:
            :class:`gantry.automations.curators.Curator`: A Curator object representing the curator
            corresponding to the provided name.
        """
        res = self._api_client.request("GET", f"/api/v1/curator/{name}", raise_for_status=True)
        return Curator.from_curator_info(self._api_client, CuratorInfo.parse_obj(res["data"]))

    @typechecked
    def delete_curator(
        self,
    ) -> "Curator":
        """
        Currently, Curators must be deleted using an instance of the Curator class.
        """
        raise NotImplementedError

    @typechecked
    def update_curator(
        self,
    ) -> "Curator":
        """
        Currently, Curators must be updated using an instance of the Curator class.
        """
        raise NotImplementedError

    @typechecked
    def enable_curator(
        self,
    ) -> "Curator":
        """
        Currently, Curators must be enabled using an instance of the Curator class.
        """
        raise NotImplementedError

    @typechecked
    def disable_curator(
        self,
    ) -> "Curator":
        """
        Currently, Curators must be disabled using an instance of the Curator class.
        """
        raise NotImplementedError

    def _get_all_curator_infos(self, application_name: Optional[str] = None) -> List["CuratorInfo"]:
        """
        Get all curators associated with an application.

        Returns:
            List[Curator]
        """
        res = self._api_client.request("GET", "/api/v1/curator", raise_for_status=True)
        res = res["data"]
        if application_name:
            return [
                CuratorInfo.parse_obj(curator)
                for curator in res
                if curator["application_name"] == application_name
            ]
        return [CuratorInfo.parse_obj(curator) for curator in res]

    def ping(self) -> bool:
        """
        Pings the API client server to check if it is alive.
        Returns True if alive, False if there is an error during ping process.
        """
        try:
            # Cannot use /healthz/* endpoints as those will be answered by nginx
            # need to use /.
            # See https://linear.app/gantry/issue/ENG-2978/revisit-ping-in-sdk
            self._api_client.request("GET", "/api/ping", raise_for_status=True)
            return True
        except Exception as e:
            logger.error(f"Error during ping: {e}")
            return False

    def ready(self) -> bool:
        """
        Checks whether the API is ready to receive traffic with the provided API Key.
        """
        try:
            self._api_client.request("GET", "/api/v1/auth-ping", raise_for_status=True)
            return True
        except Exception as e:
            logger.error(f"Error during api key check: {e}")
            return False


class Curator:
    """
    A class representing a user defined curator, where the user provides the
    selection criteria using a list of selectors, as well as other metadata
    needed to create and manage the creation jobs.
    """

    def __init__(
        self,
        name: str,
        application_name: str,
        api_client: Optional[APIClient] = None,
        id: Optional[uuid.UUID] = None,
        curated_dataset_name: Optional[str] = None,
        start_on: Optional[datetime.datetime] = None,
        curation_interval: datetime.timedelta = datetime.timedelta(days=1),
        curation_delay: datetime.timedelta = datetime.timedelta(days=0),
        curate_past_intervals: bool = True,
        created_at: Optional[datetime.datetime] = None,
        selectors: Optional[List[Selector]] = None,
    ):
        """
        A Curator defines a process that runs periodically to curate a dataset. It uses
        selectors to select rows from the dataset and write them to a curated dataset.
        The curated dataset is a new dataset that is created by the curator. The curator
        will create the curated dataset if it does not already exist, or update it if it
        does exist.f

        Instantiating a curator does not create the curator in Gantry. To create the curator,
        call `Curator.create` after instantiating the curator. Curators that have been created
        in Gantry can be instantiated as a `Curator` object using `Curator.from_curator_info`,
        although it is recommended to use the module-level `get_curator` function instead.

        Curators can be updated using `Curator.update`. This will update the curator in Gantry
        and also update the curator object in memory.

        Curators can be deleted using `Curator.delete`. This will delete the curator in Gantry
        but will not delete the curator object in memory.

        Args:
            api_client (Optional[APIClient], optional): APIClient to use. Defaults to None,
                in which case the Curator will use the global APIClient.
            id (Optional[uuid.UUID], optional): The id of the curator. Defaults to None. The
                id only exists in Gantry for curators that have been created. Typically, you
                will not need to set this. But some methods, such as
                `Curator.from_curator_info`, will set this for you.
            name str: The name of the curator.
            curated_dataset_name (Optional[str], optional): The name of the curated dataset.
                Defaults to None. If None, the curated dataset name will be the same as the
                curator name.
            application_name (str): The name of the application that the curator is running in.
            start_on (Optional[datetime.datetime], optional): The time at which the curator
                should start curating. Defaults to None. If None, the curator will start
                curating immediately, looking back one curation_interval.
            curation_interval (datetime.timedelta, optional): The interval at which the curator
                should curate. Defaults to datetime.timedelta(days=1).
            curate_past_intervals (bool, optional): Whether the curator should curate past
                intervals. Defaults to True. If True, the curator will immediately curate
                all intervals that have passed since the start_on time.
            created_at (Optional[datetime.datetime], optional): The time at which the curator
                was created. Defaults to None. The created_at time only exists in Gantry for
                curators that have been created. Typically, you will not need to set this. But
                some methods, such as `Curator.from_curator_info`, will set this for you.
            selectors (Optional[List[Selector]], optional): The selectors that the curator
                should use to curate. Defaults to None. If None, the request to create the
                curator will fail.
        """
        self._api_client: APIClient = (
            globals._API_CLIENT if api_client is None else api_client  # type: ignore
        )
        if self._api_client is None:
            raise ClientNotInitialized()
        self._id = id
        self._name = name
        self._curated_dataset_name = name if curated_dataset_name is None else curated_dataset_name
        self._application_name = application_name
        self._start_on = datetime.datetime.utcnow() if start_on is None else start_on
        self._curation_interval = curation_interval
        self._curation_delay = curation_delay
        self._curate_past_intervals = curate_past_intervals
        self._created_at = created_at
        self._selectors = [] if selectors is None else selectors

    def __repr__(self) -> str:
        return (
            f"Curator(name={self.name!r}, curated_dataset_name={self.curated_dataset_name!r}, "
            f"application_name={self.application_name!r}, start_on={self.start_on}, "
            f"curation_interval={self.curation_interval}, "
            f"curate_past_intervals={self.curate_past_intervals}, selectors={self.selectors})"
        )

    @property
    def id(self) -> uuid.UUID:
        if self._id is None:
            raise ValueError("Curator has no id. Has it been created yet?")
        return self._id

    @property
    def name(self) -> str:
        return self._name  # type: ignore

    @property
    def curated_dataset_name(self) -> str:
        return self._curated_dataset_name  # type: ignore

    @property
    def application_name(self) -> str:
        return self._application_name  # type: ignore

    @property
    def start_on(self) -> datetime.datetime:
        return self._start_on

    @property
    def curation_interval(self) -> datetime.timedelta:
        return self._curation_interval

    @property
    def curation_delay(self) -> datetime.timedelta:
        return self._curation_delay

    @property
    def curate_past_intervals(self) -> bool:
        return self._curate_past_intervals

    @property
    def created_at(self) -> datetime.datetime:
        if self._created_at is None:
            raise ValueError("Curator has not been created yet")
        return self._created_at

    @property
    def selectors(self) -> List[Selector]:
        return self._selectors

    @property
    def enabled(self) -> bool:
        """Enabling or disabling must be done via the `.enable()` or `.disable()` methods"""
        raise NotImplementedError

    def update_curation_start_on(self, start_on: datetime.datetime) -> None:
        """
        Update the curation start on time.
        Args:
            start_on (datetime.datetime): The new start on time
        """
        self._start_on = start_on

    def update_curation_interval(self, curation_interval: datetime.timedelta) -> None:
        """
        Update the curation interval.
        Args:
            curation_interval (datetime.timedelta): The new curation interval
        """
        self._curation_interval = curation_interval

    def update_curation_delay(self, curation_delay: datetime.timedelta) -> None:
        """
        Update the curation delay.
        Args:
            curation_delay (datetime.timedelta): The new curation delay
        """
        self._curation_delay = curation_delay

    def start(self) -> None:
        """
        Trigger to create the curator. This function is called from Automation.
        """
        self.create()

    def create(self, enable=True) -> "Curator":
        """
        Creates the Curator. By default, the Curator is also enabled upon creation.
        Use the `enable` parameter to change this

        Once created, the Curator will start curating the dataset according to the
        curation_interval. If curate_past_intervals is set to True, it will curate all
        past intervals and continue curating in the future. The Curator will exist in
        Gantry until it is deleted.

        The results of the curation will be stored in a dataset with the name
        curated_dataset_name. If the dataset does not exist, it will be created. If it
        does exist, it will be appended to. Curation commits will be made by the user
        with the api key used when initializing gantry services.

        The curated dataset can be accessed using the get_curated_dataset method..

        Args:
            enable (Optional[bool]): Whether to automatically enable
                the curator after creation. Defaults to True.

        Returns:
            Curator: The created curator.
        """
        create_curator_request = self._form_create_curator_request()
        result = self._create_curator(create_curator_request)
        if enable:
            result.enable()
        return result

    def update(
        self,
        new_curator_name: Optional[str] = None,
        new_curated_dataset_name: Optional[str] = None,
        new_curation_interval: Optional[datetime.timedelta] = None,
        new_selectors: Optional[List[Selector]] = None,
        create_new_dataset: bool = False,
    ) -> "Curator":
        """
        Updates the curator. At least one of the update parameters must be provided. If a parameter
        is not provided, it will not be updated.

        Args:
            new_curator_name (Optional[str], optional): The new name of the curator. Defaults to
                None.
            new_curated_dataset_name (Optional[str], optional): The name of the new dataset to
                curate to. Defaults to None.
            new_curation_interval (Optional[datetime.timedelta], optional): The new curation
                interval. Defaults to None.
            new_selectors (Optional[List[Selector]], optional): The new selectors. Defaults to
                None.
            create_new_dataset (bool, optional): Whether to create a new dataset if the requested
                name does not correspond to an existing dataset. Defaults to False.

        Raises:
            ValueError: If curator ID is not set.
            ValueError: If no update parameters are provided.

        Returns:
            Curator: The updated curator.
        """
        if self._id is None:
            raise ValueError("Curator id is required to update a curator.")
        if all(arg is None for arg in (new_curator_name, new_curation_interval, new_selectors)):
            raise ValueError("No update parameters provided")
        return self._update_curator(
            self._form_update_curator_request(
                new_curator_name=new_curator_name,
                new_curated_dataset_name=new_curated_dataset_name,
                new_curation_interval=new_curation_interval,
                new_selectors=new_selectors,
                allow_create_new_dataset=create_new_dataset,
            )
        )

    def delete(self) -> str:
        """
        Deletes the curator using the curator id. The Curator object is not deleted, but the
        curator id is reset to None, as well as the created_at time.

        Raises:
            ValueError: If curator id is not set.

        Returns:
            str: A message indicating the curator was deleted. The message also contains the
                time the curator was deleted.
        """
        if self._id is None:
            raise ValueError("Curator id is required to delete a curator.")
        res = self._api_client.request(
            "DELETE",
            f"/api/v1/curator/{self._id}",
            raise_for_status=True,
        )
        if res["response"] == "ok":
            msg = f"Curator {self._name} ({self._id}) deleted at {res['data']['deleted_at']}"
            self._id = None
            # will not be able to recreate curator with same name as long as dataset is coupled
            self._created_at = None
        return msg

    def get_curated_dataset(self) -> GantryDataset:
        """
        Gets the curated dataset associated with the Curator.

        Returns:
            GantryDataset: A GantryDataset object representing the curated dataset.
        """
        return get_dataset(self._curated_dataset_name)

    @classmethod
    def from_curator_info(cls, api_client: APIClient, curator_info: CuratorInfo) -> "Curator":
        """
        :meta private: Instantiates a Curator object from a CuratorInfo object.

        Args:
            api_client (APIClient): _description_
            curator_info (CuratorInfo): _description_

        Returns:
            Curator: _description_
        """
        return cls(api_client=api_client, **curator_info.dict())

    def _form_create_curator_request(self) -> CreateCuratorRequest:
        """
        Forms a create curator request model.

        Returns:
            CreateCuratorRequest: Create request model.
        """
        return CreateCuratorRequest(
            name=self._name,
            curated_dataset_name=self._curated_dataset_name,
            application_name=self._application_name,
            start_on=self._start_on,
            curation_interval=self._curation_interval,
            curation_delay=self._curation_delay,
            curate_past_intervals=self._curate_past_intervals,
            selectors=self._selectors,
        )

    def _create_curator(
        self,
        create_curator_request: CreateCuratorRequest,
    ) -> "Curator":
        """
        Makes the API call to create a curator.

        Args:
            create_curator_request (CreateCuratorRequest): Create reqest model.

        Returns:
            Curator: The created curator.
        """

        res = self._api_client.request(
            "POST",
            "/api/v1/curator",
            json=json.loads(create_curator_request.json()),
            raise_for_status=True,
        )

        curator_info = CuratorInfo.parse_obj(res["data"])
        self._id = curator_info.id
        self._name = curator_info.name
        self._curated_dataset_name = curator_info.curated_dataset_name
        self._application_name = curator_info.application_name
        self._start_on = curator_info.start_on
        self._curation_interval = curator_info.curation_interval
        self._curate_past_intervals = curator_info.curate_past_intervals
        self._created_at = curator_info.created_at
        self._selectors = curator_info.selectors
        return self

    def _form_update_curator_request(
        self,
        new_curator_name: Optional[str] = None,
        new_curated_dataset_name: Optional[str] = None,
        new_curation_interval: Optional[datetime.timedelta] = None,
        new_selectors: Optional[List[Selector]] = None,
        allow_create_new_dataset: bool = False,
    ) -> UpdateCuratorRequest:
        """
        Instantiates a valid update request model.

        Args:
            new_curator_name (Optional[str], optional): The new name of the curator. Defaults to
                None.
            new_curation_interval (Optional[datetime.timedelta], optional): The new curation
                interval. Defaults to None.
            new_selectors (Optional[List[Selector]], optional): The new selectors. Defaults to
                None.

        Returns:
            UpdateCuratorRequest: Valid update request model.
        """
        return UpdateCuratorRequest(
            name=self._name,
            new_name=new_curator_name,
            new_curated_dataset_name=new_curated_dataset_name,
            curation_interval=new_curation_interval,
            selectors=new_selectors,
            allow_create_new_dataset=allow_create_new_dataset,
        )

    def _update_curator(self, update_curator_request: UpdateCuratorRequest) -> "Curator":
        """
        Issues the request to update the curator using the API client and an update request.

        Args:
            update_curator_request (UpdateCuratorRequest): Update request model.

        Returns:
            Curator: the updated curator.
        """
        res = self._api_client.request(
            "PATCH",
            "/api/v1/curator",
            json=json.loads(update_curator_request.json()),
            raise_for_status=True,
        )
        curator_info = CuratorInfo.parse_obj(res["data"])
        self._name = curator_info.name
        self._curated_dataset_name = curator_info.curated_dataset_name
        self._curation_interval = curator_info.curation_interval
        self._selectors = curator_info.selectors
        return self

    def enable(self) -> str:
        """
        Enable Curator - Gantry will resume curating data using this Curator.
        Intervals that have elapsed since the curator was disabled will be backfilled.
        """
        payload = EnableCuratorRequest(name=self.name, enable=True)
        try:
            res = self._api_client.request(
                "PATCH",
                "/api/v1/curator/enable",
                json=json.loads(payload.json()),
                raise_for_status=True,
            )
        except GantryRequestException as e:
            raise CuratorNotFoundError(
                f"Could not enable {self.name}, did you create the curator?"
            ) from e
        return res["data"]

    def disable(self) -> str:
        """Disable Curator - Gantry will stop curating data using this Curator."""
        payload = EnableCuratorRequest(name=self.name, enable=False)
        try:
            res = self._api_client.request(
                "PATCH",
                "/api/v1/curator/enable",
                json=json.loads(payload.json()),
                raise_for_status=True,
            )
        except GantryRequestException as e:
            raise CuratorNotFoundError(
                f"Could not disable {self.name}, did you create the curator?"
            ) from e
        return res["data"]



import datetime

import pytest
import responses
from responses import matchers

from gantry.api_client import APIClient
from gantry.automations.curators import Curator
from gantry.automations.curators.curators import CuratorClient
from gantry.automations.curators.selectors import Selector

from .conftest import HOST


@pytest.fixture
def test_api_client():
    return APIClient(origin=HOST)


@pytest.fixture
def test_client(test_api_client):
    return CuratorClient(api_client=test_api_client)


@pytest.mark.parametrize(
    ("application_name", "expected_len"),
    [(None, 2), ("test_curator_app", 1), ("different_app_name", 0)],
)
def test_get_curators(test_client, test_curators, application_name, expected_len):
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/curator",
            json={
                "response": "ok",
                "data": test_curators,
            },
            headers={"Content-Type": "application/json"},
        )
        curators = test_client.get_all_curators(application_name=application_name)
        assert len(curators) == expected_len


@pytest.mark.parametrize(
    ("application_name", "expected_len"),
    [(None, 2), ("test_curator_app", 1), ("different_app_name", 0)],
)
def test_list_curators(test_client, test_curators, application_name, expected_len):
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/curator",
            json={
                "response": "ok",
                "data": test_curators,
            },
            headers={"Content-Type": "application/json"},
        )
        curator_names = test_client.list_curators(application_name=application_name)
        assert len(curator_names) == expected_len


def test_get_curator(test_client, test_curator):
    with responses.RequestsMock() as resp:
        resp.add(
            resp.GET,
            f"{HOST}/api/v1/curator/test_curator_name",
            json={
                "response": "ok",
                "data": test_curator,
            },
            headers={"Content-Type": "application/json"},
        )
        _ = test_client.get_curator(name="test_curator_name")


def test_create_curator_disabled(test_api_client, test_curators):
    curator = Curator(**test_curators[0], api_client=test_api_client)

    with responses.RequestsMock() as resp:
        resp.add(
            resp.POST,
            f"{HOST}/api/v1/curator",
            json={
                "response": "ok",
                "data": test_curators[1],
            },
            headers={"Content-Type": "application/json"},
        )
        assert curator.create(enable=False) is curator

    assert str(curator.id) == test_curators[1]["id"]
    assert curator.name == test_curators[1]["name"]
    assert curator.curated_dataset_name == test_curators[1]["curated_dataset_name"]
    assert curator.application_name == test_curators[1]["application_name"]
    assert curator.start_on.isoformat() == test_curators[1]["start_on"]
    assert str(curator.curation_interval) == test_curators[1]["curation_interval"]
    assert curator.curate_past_intervals == test_curators[1]["curate_past_intervals"]
    assert curator.created_at.isoformat() == test_curators[1]["created_at"]
    assert curator.selectors == [Selector(**info) for info in test_curators[1]["selectors"]]


def test_update_curator(test_api_client, test_curator):
    curator = Curator(**test_curator, api_client=test_api_client)

    update_info = {
        "name": "updated_name",
        "curated_dataset_name": "updated_dataset_name",
        "curation_interval": str(datetime.timedelta(hours=5)),
        "selectors": [Selector(limit=123).dict()],
    }

    updated_json = {
        key: test_curator[key] if key not in update_info else update_info[key]
        for key in test_curator
    }

    with responses.RequestsMock() as resp:
        resp.add(
            resp.PATCH,
            f"{HOST}/api/v1/curator",
            json={
                "response": "ok",
                "data": updated_json,
            },
            headers={"Content-Type": "application/json"},
        )

        assert (
            curator.update(
                new_curator_name=update_info["name"],
                new_curated_dataset_name=update_info["curated_dataset_name"],
                new_curation_interval=update_info["curation_interval"],
                new_selectors=update_info["selectors"],
            )
            is curator
        )

    assert curator.name == update_info["name"]
    assert str(curator.curation_interval) == update_info["curation_interval"]
    assert curator.selectors == update_info["selectors"]
    assert curator.curated_dataset_name == update_info["curated_dataset_name"]


def test_delete_curator(test_api_client, test_curator):
    curator = Curator(**test_curator, api_client=test_api_client)

    with responses.RequestsMock() as resp:
        resp.add(
            resp.DELETE,
            f"{HOST}/api/v1/curator/{test_curator['id']}",
            json={
                "response": "ok",
                "data": {
                    "curated_dataset_name": test_curator["name"],
                    "id": test_curator["id"],
                    "application_name": test_curator["application_name"],
                    "deleted_at": str(datetime.datetime(2022, 1, 1)),
                },
            },
            headers={"Content-Type": "application/json"},
        )

        curator_info = f"{test_curator['name']} ({test_curator['id']})"
        assert curator.delete() == f"Curator {curator_info} deleted at 2022-01-01 00:00:00"

    with pytest.raises(ValueError):
        _ = curator.id
    with pytest.raises(ValueError):
        _ = curator.created_at


def test_enable(test_api_client, test_curator):
    curator = Curator(**test_curator, api_client=test_api_client)

    with responses.RequestsMock() as resp:
        resp.add(
            resp.PATCH,
            f"{HOST}/api/v1/curator/enable",
            json={
                "response": "ok",
                "data": "response_data",
            },
            headers={"Content-Type": "application/json"},
            match=[matchers.json_params_matcher({"name": curator.name, "enable": True})],
        )
        assert curator.enable() == "response_data"


def test_disable(test_api_client, test_curator):
    curator = Curator(**test_curator, api_client=test_api_client)

    with responses.RequestsMock() as resp:
        resp.add(
            resp.PATCH,
            f"{HOST}/api/v1/curator/enable",
            json={
                "response": "ok",
                "data": "response_data",
            },
            headers={"Content-Type": "application/json"},
            match=[matchers.json_params_matcher({"name": curator.name, "enable": False})],
        )
        assert curator.disable() == "response_data"


def test_create_curator_default_enabled(test_api_client, test_curators):
    curator = Curator(**test_curators[0], api_client=test_api_client)

    with responses.RequestsMock() as resp:
        resp.add(
            resp.POST,
            f"{HOST}/api/v1/curator",
            json={
                "response": "ok",
                "data": test_curators[1],
            },
            headers={"Content-Type": "application/json"},
        )
        resp.add(
            resp.PATCH,
            f"{HOST}/api/v1/curator/enable",
            json={
                "response": "ok",
                "data": "response_data",
            },
            headers={"Content-Type": "application/json"},
            match=[
                matchers.json_params_matcher({"name": test_curators[1]["name"], "enable": True})
            ],
        )
        assert curator.create() is curator

    assert str(curator.id) == test_curators[1]["id"]
    assert curator.name == test_curators[1]["name"]



from gantry.query import globals
from gantry.query.client import GantryQuery
from gantry.query.globals import _query_alias, validate_init


@_query_alias
def accuracy_score(*args, **kwargs):
    validate_init()
    assert isinstance(globals._Query, GantryQuery)  # mypy bug- mypy/issues/4805
    return globals._Query.metric.accuracy_score(*args, **kwargs)


@_query_alias
def mean_squared_error(*args, **kwargs):
    validate_init()
    assert isinstance(globals._Query, GantryQuery)  # mypy bug- mypy/issues/4805
    return globals._Query.metric.mean_squared_error(*args, **kwargs)


@_query_alias
def confusion_matrix(*args, **kwargs):
    validate_init()
    assert isinstance(globals._Query, GantryQuery)  # mypy bug- mypy/issues/4805
    return globals._Query.metric.confusion_matrix(*args, **kwargs)


@_query_alias
def f1_score(*args, **kwargs):
    validate_init()
    assert isinstance(globals._Query, GantryQuery)  # mypy bug- mypy/issues/4805
    return globals._Query.metric.f1_score(*args, **kwargs)


@_query_alias
def r2_score(*args, **kwargs):
    validate_init()
    assert isinstance(globals._Query, GantryQuery)  # mypy bug- mypy/issues/4805
    return globals._Query.metric.r2_score(*args, **kwargs)


@_query_alias
def precision_score(*args, **kwargs):
    validate_init()
    assert isinstance(globals._Query, GantryQuery)  # mypy bug- mypy/issues/4805
    return globals._Query.metric.precision_score(*args, **kwargs)


@_query_alias
def recall_score(*args, **kwargs):
    validate_init()
    assert isinstance(globals._Query, GantryQuery)  # mypy bug- mypy/issues/4805
    return globals._Query.metric.recall_score(*args, **kwargs)


@_query_alias
def roc_auc_score(*args, **kwargs):
    validate_init()
    assert isinstance(globals._Query, GantryQuery)  # mypy bug- mypy/issues/4805
    return globals._Query.metric.roc_auc_score(*args, **kwargs)


@_query_alias
def percent_null(*args, **kwargs):
    validate_init()
    assert isinstance(globals._Query, GantryQuery)  # mypy bug- mypy/issues/4805
    return globals._Query.metric.percent_null(*args, **kwargs)


@_query_alias
def percent_true(*args, **kwargs):
    validate_init()
    assert isinstance(globals._Query, GantryQuery)  # mypy bug- mypy/issues/4805
    return globals._Query.metric.percent_true(*args, **kwargs)


@_query_alias
def percent_false(*args, **kwargs):
    validate_init()
    assert isinstance(globals._Query, GantryQuery)  # mypy bug- mypy/issues/4805
    return globals._Query.metric.percent_false(*args, **kwargs)


@_query_alias
def percent_true_not_null(*args, **kwargs):
    validate_init()
    assert isinstance(globals._Query, GantryQuery)  # mypy bug- mypy/issues/4805
    return globals._Query.metric.percent_true_not_null(*args, **kwargs)


@_query_alias
def percent_false_not_null(*args, **kwargs):
    validate_init()
    assert isinstance(globals._Query, GantryQuery)  # mypy bug- mypy/issues/4805
    return globals._Query.metric.percent_false_not_null(*args, **kwargs)
